from __future__ import annotations

import logging
import uuid
from http import HTTPStatus
from typing import TYPE_CHECKING

from aiohttp import web

from ai.backend.common.api_handlers import (
    APIResponse,
    BodyParam,
    api_handler,
)
from ai.backend.common.dto.storage.request import (
    ReservoirImportModelsReq,
)
from ai.backend.common.dto.storage.response import (
    ReservoirImportModelsResponse,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.storage.services.artifacts.reservoir import (
    ReservoirService,
    ReservoirServiceArgs,
    create_reservoir_import_pipeline,
)

from ....utils import log_client_api_entry

if TYPE_CHECKING:
    from ....context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ReservoirRegistryAPIHandler:
    _reservoir_service: ReservoirService

    def __init__(
        self,
        reservoir_service: ReservoirService,
    ) -> None:
        self._reservoir_service = reservoir_service

    @api_handler
    async def import_models(
        self,
        body: BodyParam[ReservoirImportModelsReq],
    ) -> APIResponse:
        """
        Import multiple models to storage in batch.
        """
        await log_client_api_entry(log, "import_models", None)

        # Create import pipeline based on storage step mappings
        pipeline = create_reservoir_import_pipeline(
            storage_pool=self._reservoir_service._storage_pool,
            registry_configs=self._reservoir_service._reservoir_registry_configs,
            storage_step_mappings=body.parsed.storage_step_mappings,
            transfer_manager=self._reservoir_service._transfer_manager,
            artifact_verifier_ctx=self._reservoir_service._artifact_verifier_ctx,
            event_producer=self._reservoir_service._event_producer,
            manager_client_pool=self._reservoir_service._manager_client_pool,
            redis_client=self._reservoir_service._redis_client,
        )

        task_id = await self._reservoir_service.import_models_batch(
            registry_name=body.parsed.registry_name,
            models=body.parsed.models,
            storage_step_mappings=body.parsed.storage_step_mappings,
            pipeline=pipeline,
            artifact_revision_ids=[
                uuid.UUID(rev_id) for rev_id in body.parsed.artifact_revision_ids
            ],
        )

        return APIResponse.build(
            status_code=HTTPStatus.ACCEPTED,
            response_model=ReservoirImportModelsResponse(task_id=task_id),
        )


def create_app(ctx: RootContext) -> web.Application:
    app = web.Application()
    app["ctx"] = ctx
    app["prefix"] = "v1/registries/reservoir"

    reservoir_service = ReservoirService(
        ReservoirServiceArgs(
            background_task_manager=ctx.background_task_manager,
            event_producer=ctx.event_producer,
            storage_pool=ctx.storage_pool,
            reservoir_registry_configs=ctx.manager_client_pool.registry_configs,
            artifact_verifier_ctx=ctx.artifact_verifier_ctx,
            manager_client_pool=ctx.manager_client_pool,
            redis_client=ctx.valkey_artifact_client,
        )
    )
    reservoir_api_handler = ReservoirRegistryAPIHandler(
        reservoir_service=reservoir_service,
    )

    app.router.add_route("POST", "/import", reservoir_api_handler.import_models)

    return app



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator, Optional

from aiohttp import web

from ai.backend.common.plugin import AbstractPlugin, BasePluginContext
from ai.backend.storage.services.artifacts.types import ImportStepContext

if TYPE_CHECKING:
    from .api.types import CORSOptions, WebMiddleware
    from .volumes.abc import AbstractVolume


class AbstractStoragePlugin(AbstractPlugin, metaclass=ABCMeta):
    @abstractmethod
    def get_volume_class(
        self,
    ) -> type[AbstractVolume]:
        raise NotImplementedError


@dataclass
class VerifierPluginResult:
    scanned_count: int
    infected_count: int
    metadata: dict[str, Any]


class AbstractArtifactVerifierPlugin(AbstractPlugin, metaclass=ABCMeta):
    @abstractmethod
    async def verify(self, artifact_path: Path, context: ImportStepContext) -> VerifierPluginResult:
        raise NotImplementedError


class StoragePluginContext(BasePluginContext[AbstractStoragePlugin]):
    plugin_group = "backendai_storage_v10"

    @classmethod
    def discover_plugins(
        cls,
        plugin_group: str,
        allowlist: Optional[set[str]] = None,
        blocklist: Optional[set[str]] = None,
    ) -> Iterator[tuple[str, type[AbstractStoragePlugin]]]:
        scanned_plugins = [*super().discover_plugins(plugin_group, allowlist, blocklist)]
        yield from scanned_plugins


class StorageArtifactVerifierPluginContext(BasePluginContext[AbstractArtifactVerifierPlugin]):
    plugin_group = "backendai_storage_artifact_verifier_v1"

    @classmethod
    def discover_plugins(
        cls,
        plugin_group: str,
        allowlist: Optional[set[str]] = None,
        blocklist: Optional[set[str]] = None,
    ) -> Iterator[tuple[str, type[AbstractArtifactVerifierPlugin]]]:
        scanned_plugins = [*super().discover_plugins(plugin_group, allowlist, blocklist)]
        yield from scanned_plugins


class StorageManagerWebappPlugin(AbstractPlugin, metaclass=ABCMeta):
    @abstractmethod
    async def create_app(
        self,
        cors_options: CORSOptions,
    ) -> tuple[web.Application, list[WebMiddleware]]:
        raise NotImplementedError


class StorageManagerWebappPluginContext(BasePluginContext[StorageManagerWebappPlugin]):
    plugin_group = "backendai_storage_manager_webapp_v10"


class StorageClientWebappPlugin(AbstractPlugin, metaclass=ABCMeta):
    @abstractmethod
    async def create_app(
        self,
        cors_options: CORSOptions,
    ) -> tuple[web.Application, list[WebMiddleware]]:
        raise NotImplementedError


class StorageClientWebappPluginContext(BasePluginContext[StorageClientWebappPlugin]):
    plugin_group = "backendai_storage_client_webapp_v10"



import enum
from typing import Optional

from pydantic import BaseModel, ConfigDict

from .placement_objective import PlacementObjective
from .uoid import UOID


class Priority(enum.StrEnum):
    LOW = "LOW"
    MEDIUM_LOW = "MEDIUM_LOW"
    MEDIUM = "MEDIUM"
    MEDIUM_HIGH = "MEDIUM_HIGH"
    HIGH = "HIGH"


class SimpleObjective(BaseModel):
    model_config = ConfigDict(extra="allow")

    uoid: UOID
    name: str
    zombie: bool
    hidden: bool
    internalId: int


class AppliedObjective(BaseModel):
    applicability: str
    id: int


class Objective(BaseModel):
    model_config = ConfigDict(extra="allow")

    uoid: UOID
    name: str
    modified: int  # timestamp
    priority: Priority
    appliedObjectives: list[AppliedObjective]
    basic: bool
    internalId: int
    hidden: bool
    placementObjective: Optional[PlacementObjective] = None



import uuid

from pydantic import BaseModel, ConfigDict

from .uoid import UOID


class Location(BaseModel):
    model_config = ConfigDict(extra="allow")

    uoid: UOID
    modified: int  # timestamp
    excludeFromObjectiveUuid: uuid.UUID
    confineToObjectiveUuid: uuid.UUID
    placeOnObjectiveUuid: uuid.UUID


class LocationList(BaseModel):
    id: int
    placeOn: list[Location]



from __future__ import annotations

import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

from .schema.object_type import ObjectType


@dataclass
class CreateShareParams:
    name: str
    path: Path
    share_size_limit: Optional[int] = None
    create_path: bool = True
    validate_only: bool = False

    def query(self) -> dict[str, str]:
        return {
            "create-path": "true" if self.create_path else "false",
            "validate-only": "true" if self.validate_only else "false",
        }

    def body(self) -> dict[str, Any]:
        data: dict[str, Any] = {
            "name": self.name,
            "path": str(self.path),
        }
        if self.share_size_limit is not None:
            data["shareSizeLimit"] = self.share_size_limit
        return data


@dataclass
class GetShareParams:
    name: str

    def query(self) -> dict[str, str]:
        return {"spec": f"name=eq={self.name}"}


@dataclass
class ClusterMetricParams:
    site_id: uuid.UUID
    preceding_duration: str  # "7d", "24h", "60m"
    interval_duration: str  # "1d", "1h", "5m"

    # Available object_type values : CLUSTER, SHARE, STORAGE_VOLUME
    object_type: ObjectType = ObjectType.CLUSTER

    def query(self) -> dict[str, str]:
        return {
            "precedingDuration": self.preceding_duration,
            "intervalDuration": self.interval_duration,
        }



"""
Process/subprocess execution exceptions.
"""

from __future__ import annotations

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)

from .base import ExternalStorageServiceError


class SubprocessStdoutNotAvailableError(BackendAIError, web.HTTPInternalServerError):
    """Raised when subprocess stdout is not available."""

    error_type = "https://api.backend.ai/probs/storage/subprocess/stdout-unavailable"
    error_title = "Subprocess Stdout Not Available"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.EXECUTE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class QuotaCommandFailedError(BackendAIError, web.HTTPInternalServerError):
    """Raised when a quota command fails."""

    error_type = "https://api.backend.ai/probs/storage/quota/command-failed"
    error_title = "Quota Command Failed"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.QUOTA_SCOPE,
            operation=ErrorOperation.EXECUTE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class CephNotInstalledError(BackendAIError, web.HTTPServiceUnavailable):
    """Raised when Ceph is not installed."""

    error_type = "https://api.backend.ai/probs/storage/ceph/not-installed"
    error_title = "Ceph Not Installed"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.SETUP,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class PureStorageCommandFailedError(BackendAIError, web.HTTPInternalServerError):
    """Raised when a PureStorage command fails."""

    error_type = "https://api.backend.ai/probs/storage/purestorage/command-failed"
    error_title = "PureStorage Command Failed"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.EXECUTE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class NetAppClientError(ExternalStorageServiceError, web.HTTPServiceUnavailable):
    """Raised when a NetApp API call fails."""

    error_type = "https://api.backend.ai/probs/storage/netapp/api-error"
    error_title = "NetApp API Error"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.REQUEST,
            error_detail=ErrorDetail.UNREACHABLE,
        )


class NetAppQTreeNotFoundError(BackendAIError, web.HTTPNotFound):
    """Raised when a NetApp qtree is not found."""

    error_type = "https://api.backend.ai/probs/storage/netapp/qtree-not-found"
    error_title = "NetApp QTree Not Found"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.QUOTA_SCOPE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class DDNCommandFailedError(BackendAIError, web.HTTPInternalServerError):
    """Raised when a DDN lfs command fails."""

    error_type = "https://api.backend.ai/probs/storage/ddn/command-failed"
    error_title = "DDN Command Failed"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.EXECUTE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class MetricNotFoundError(BackendAIError, web.HTTPNotFound):
    """Raised when a metric is not found."""

    error_type = "https://api.backend.ai/probs/storage/metric-not-found"
    error_title = "Metric Not Found"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )



from __future__ import annotations

from .composer import InfrastructureComposer, InfrastructureComposerInput, InfrastructureResources
from .etcd import EtcdProvider
from .redis import RedisProvider, StorageProxyValkeyClients

__all__ = [
    "EtcdProvider",
    "InfrastructureComposer",
    "InfrastructureComposerInput",
    "InfrastructureResources",
    "RedisProvider",
    "StorageProxyValkeyClients",
]



import enum

from pydantic import BaseModel, ConfigDict

from .address import IPAddress, QualifiedAddress
from .capacity import Capacity
from .fstype import FsType
from .node import Node
from .service_state import ServiceState
from .uoid import UOID


class UsageType(enum.StrEnum):
    DS = "DS"
    OTHER = "OTHER"


class LogicalVolume(BaseModel):
    model_config = ConfigDict(extra="allow")

    uoid: UOID
    name: str
    modified: int  # timestamp
    quotaIsSet: bool
    serviceState: ServiceState
    addresses: list[QualifiedAddress]
    exportPath: str  # NFS export path
    aliases: list[str]
    fsType: FsType
    capacity: Capacity
    ipAddresses: list[IPAddress]
    node: Node
    usageType: UsageType



from __future__ import annotations

import logging
import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Any,
    ClassVar,
    Mapping,
    Optional,
    override,
)

from yarl import URL

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.dispatcher import EventDispatcher, EventProducer
from ai.backend.common.events.event_types.volume.broadcast import DoVolumeMountEvent
from ai.backend.common.types import QuotaConfig, QuotaScopeID
from ai.backend.logging import BraceStyleAdapter
from ai.backend.storage.watcher import WatcherClient

from ....types import CapacityUsage, QuotaUsage
from ...abc import (
    CAP_QUOTA,
    CAP_VFOLDER,
    AbstractQuotaModel,
)
from ...vfs import BaseQuotaModel
from ..client import HammerspaceAPIClient
from ..exception import (
    AuthenticationError,
    ConfigurationError,
)
from ..request import ClusterMetricParams, CreateShareParams, GetShareParams
from ..schema.metric import ValidClusterMetricRow
from ..schema.share import Share, SimpleShare
from ..types import APIConnectionInfo, SSLConfig
from .base import BaseHammerspaceVolume

METRIC_PRECEDING_DURATION = "1h"
METRIC_INTERVAL_DURATION = "5m"


log = BraceStyleAdapter(logging.getLogger(__name__))


@dataclass
class HammerspaceQuotaModelBaseArgs:
    event_producer: EventProducer
    mount_source: str
    mount_target_path: Path
    share_query_retry: int
    share_query_wait_sec: int


class HammerspaceQuotaModel(BaseQuotaModel):
    _client: HammerspaceAPIClient
    _event_producer: EventProducer

    def __init__(
        self,
        base_args: HammerspaceQuotaModelBaseArgs,
        client: HammerspaceAPIClient,
    ) -> None:
        super().__init__(mount_path=base_args.mount_target_path)
        self._client = client
        self._event_producer = base_args.event_producer
        self._mount_source = base_args.mount_source
        # `_mount_target_path` is not used becase:
        # Agents has `mount_path` config which is the base path for all mounts
        self._mount_target_path = base_args.mount_target_path

        self._share_query_retry = base_args.share_query_retry
        self._share_query_wait_sec = base_args.share_query_wait_sec

    def _get_share_name(self, quota_scope_id: QuotaScopeID) -> str:
        # Not allowed to name Share with the following characters:
        # \"/\\[]:|<>+;,?*=
        return quota_scope_id.pathname

    def _get_share_path(self, quota_scope_id: QuotaScopeID) -> Path:
        return Path("/", quota_scope_id.pathname)

    def _get_mount_source(self, share: Share | SimpleShare) -> str:
        return f"{self._mount_source}:{share.path}"

    @override
    async def create_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        options: Optional[QuotaConfig] = None,
        extra_args: Optional[dict[str, Any]] = None,
    ) -> None:
        name = self._get_share_name(quota_scope_id)
        path = self._get_share_path(quota_scope_id)
        share_size_limit = options.limit_bytes if options is not None else None

        share = await self._client.create_share(
            CreateShareParams(
                name=name,
                path=path,
                share_size_limit=share_size_limit,
                create_path=True,
                validate_only=False,
            ),
            retry=self._share_query_retry,
            wait_sec=self._share_query_wait_sec,
        )
        dir_name = str(self.mangle_qspath(quota_scope_id))
        await self._event_producer.broadcast_event(
            DoVolumeMountEvent(
                dir_name=dir_name,
                volume_backend_name="hammerspace",
                fs_location=self._get_mount_source(share),
                quota_scope_id=quota_scope_id,
                edit_fstab=True,
            ),
            None,
        )

    @override
    async def describe_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> Optional[QuotaUsage]:
        name = self._get_share_name(quota_scope_id)
        share = await self._client.get_share(
            GetShareParams(name=name),
        )
        if share is None:
            return None
        return QuotaUsage(
            used_bytes=share.space.used,
            limit_bytes=share.space.total,
        )

    @override
    async def update_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        config: QuotaConfig,
    ) -> None:
        # Hammerspace does not support updating shares.
        # TODO: Raise not implemented error and handle it in Manager
        pass

    @override
    async def delete_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> None:
        pass


class HammerspaceVolume(BaseHammerspaceVolume):
    name: ClassVar[str] = "hammerspace"

    def __init__(
        self,
        local_config: Mapping[str, Any],
        mount_path: Path,
        *,
        etcd: AsyncEtcd,
        event_dispatcher: EventDispatcher,
        event_producer: EventProducer,
        watcher: Optional[WatcherClient] = None,
        options: Optional[Mapping[str, Any]] = None,
    ) -> None:
        self.local_config = local_config
        self.mount_path = mount_path
        self.config = options or {}
        self.etcd = etcd
        self.event_dispatcher = event_dispatcher
        self.event_producer = event_producer
        self.watcher = watcher

        address = self.config.get("address")
        if address is None:
            raise ConfigurationError("Hammerspace volume requires 'address' in options")
        username = self.config.get("username")
        if username is None:
            raise ConfigurationError("Hammerspace volume requires 'username' in options")
        password = self.config.get("password")
        if password is None:
            raise ConfigurationError("Hammerspace volume requires 'password' in options")

        mount_source = self.config.get("mount_source")
        if mount_source is None:
            raise ConfigurationError("Hammerspace volume requires 'mount_source' in options")
        self._mount_source = mount_source

        share_query_retry = self.config.get("share_query_retry", 5)
        share_query_wait_sec = self.config.get("share_query_wait_sec", 1)
        self._share_query_retry = int(share_query_retry)
        self._share_query_wait_sec = int(share_query_wait_sec)

        ssl_enabled = self.config.get("ssl_enabled", False)
        raw_ssl_config = self.config.get("ssl_config", None)
        ssl_config: Optional[SSLConfig] = None
        if raw_ssl_config is not None:
            ssl_config = SSLConfig.model_validate(raw_ssl_config)
        self._connection_info = APIConnectionInfo(
            address=URL(address),
            username=username,
            password=password,
            ssl_enabled=ssl_enabled,
            ssl_config=ssl_config,
        )

    @override
    async def init(self) -> None:
        self._client = HammerspaceAPIClient(self._connection_info)
        try:
            await self._client.try_login()
        except AuthenticationError as e:
            raise ConfigurationError(
                "Failed to authenticate to Hammerspace. "
                f"Please check your user account. (username: {self._client._connection_info.username})"
            ) from e

        await super().init()

    @override
    async def create_quota_model(self) -> AbstractQuotaModel:
        base_args = HammerspaceQuotaModelBaseArgs(
            event_producer=self.event_producer,
            mount_source=self._mount_source,
            mount_target_path=self.mount_path,
            share_query_retry=self._share_query_retry,
            share_query_wait_sec=self._share_query_wait_sec,
        )
        quota_model = HammerspaceQuotaModel(
            base_args,
            self._client,
        )
        return quota_model

    @override
    async def get_capabilities(self) -> frozenset[str]:
        return frozenset([CAP_VFOLDER, CAP_QUOTA])

    async def _get_site_id(self) -> Optional[uuid.UUID]:
        sites = await self._client.get_sites()
        if not sites:
            return None
        site = sites[0]
        return site.uoid.uuid

    async def _get_site_metric_series_row(
        self, site_id: uuid.UUID
    ) -> Optional[ValidClusterMetricRow]:
        metrics = await self._client.get_cluster_metrics(
            params=ClusterMetricParams(
                site_id=site_id,
                preceding_duration=METRIC_PRECEDING_DURATION,
                interval_duration=METRIC_INTERVAL_DURATION,
            )
        )
        if not metrics.series:
            return None

        metric_series = metrics.series[0]
        if not metric_series.rows:
            return None

        sorted_rows = sorted(metric_series.rows, key=lambda r: r.time, reverse=True)
        for row in sorted_rows:
            if (valid_row := row.parse()) is not None:
                return valid_row
        return None

    @override
    async def get_fs_usage(self) -> CapacityUsage:
        site_id = await self._get_site_id()
        if site_id is None:
            log.warning("No sites found in the Hammerspace cluster")
            return CapacityUsage(
                capacity_bytes=-1,
                used_bytes=-1,
            )

        valid_row = await self._get_site_metric_series_row(site_id)
        if valid_row is not None:
            return CapacityUsage(
                capacity_bytes=valid_row.total,
                used_bytes=valid_row.used,
            )

        log.warning("No valid metric series rows found in the site (id:{})", site_id)
        return CapacityUsage(
            capacity_bytes=-1,
            used_bytes=-1,
        )



from __future__ import annotations

import logging
from collections.abc import MutableMapping
from contextlib import asynccontextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import (
    AsyncIterator,
    Final,
    Mapping,
)

import aiohttp_cors

from ai.backend.common.bgtask.bgtask import BackgroundTaskManager
from ai.backend.common.clients.valkey_client.valkey_artifact.client import (
    ValkeyArtifactDownloadTrackingClient,
)
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.dispatcher import (
    EventDispatcher,
    EventProducer,
)
from ai.backend.common.health_checker.probe import HealthProbe
from ai.backend.common.metrics.metric import CommonMetricRegistry
from ai.backend.logging import BraceStyleAdapter

from .client.manager import ManagerHTTPClientPool
from .config.unified import StorageProxyUnifiedConfig
from .context_types import ArtifactVerifierContext
from .errors import InvalidVolumeError
from .plugin import (
    StorageArtifactVerifierPluginContext,
)
from .services.service import VolumeService
from .storages.storage_pool import StoragePool
from .types import VolumeInfo
from .volumes.abc import AbstractVolume
from .volumes.cephfs import CephFSVolume
from .volumes.ddn import EXAScalerFSVolume
from .volumes.dellemc import DellEMCOneFSVolume
from .volumes.gpfs import GPFSVolume
from .volumes.hammerspace.volume.base import BaseHammerspaceVolume
from .volumes.hammerspace.volume.extended import HammerspaceVolume
from .volumes.netapp import NetAppVolume
from .volumes.noop import NoopVolume
from .volumes.pool import VolumePool
from .volumes.purestorage import FlashBladeVolume
from .volumes.vast import VASTVolume
from .volumes.vfs import BaseVolume
from .volumes.weka import WekaVolume
from .volumes.xfs import XfsVolume
from .watcher import WatcherClient

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

EVENT_DISPATCHER_CONSUMER_GROUP: Final = "storage-proxy"

DEFAULT_BACKENDS: Mapping[str, type[AbstractVolume]] = {
    FlashBladeVolume.name: FlashBladeVolume,
    BaseVolume.name: BaseVolume,
    XfsVolume.name: XfsVolume,
    NetAppVolume.name: NetAppVolume,
    # NOTE: Dell EMC has two different storage: PowerStore and PowerScale (OneFS).
    #       We support the latter only for now.
    DellEMCOneFSVolume.name: DellEMCOneFSVolume,
    WekaVolume.name: WekaVolume,
    GPFSVolume.name: GPFSVolume,  # IBM SpectrumScale or GPFS
    "spectrumscale": GPFSVolume,  # IBM SpectrumScale or GPFS
    CephFSVolume.name: CephFSVolume,
    VASTVolume.name: VASTVolume,
    EXAScalerFSVolume.name: EXAScalerFSVolume,
    NoopVolume.name: NoopVolume,
    HammerspaceVolume.name: HammerspaceVolume,
    BaseHammerspaceVolume.name: BaseHammerspaceVolume,
}


class ServiceContext:
    volume_service: VolumeService

    def __init__(
        self,
        service: VolumeService,
    ) -> None:
        self.volume_service = service


@dataclass(slots=True)
class RootContext:
    # configuration context
    pid: int
    pidx: int
    node_id: str
    local_config: StorageProxyUnifiedConfig
    etcd: AsyncEtcd

    # internal services
    volume_pool: VolumePool
    storage_pool: StoragePool
    event_producer: EventProducer
    event_dispatcher: EventDispatcher
    watcher: WatcherClient | None
    metric_registry: CommonMetricRegistry
    background_task_manager: BackgroundTaskManager
    cors_options: Mapping[str, aiohttp_cors.ResourceOptions]
    manager_client_pool: ManagerHTTPClientPool
    valkey_artifact_client: ValkeyArtifactDownloadTrackingClient
    health_probe: HealthProbe

    # volume backend states
    backends: MutableMapping[str, type[AbstractVolume]]
    volumes: MutableMapping[str, AbstractVolume]
    artifact_verifier_ctx: ArtifactVerifierContext

    async def init_storage_artifact_verifier_plugin(self) -> None:
        plugin_ctx = StorageArtifactVerifierPluginContext(self.etcd, self.local_config.model_dump())
        await plugin_ctx.init()
        plugins = {}
        for plugin_name, plugin_instance in plugin_ctx.plugins.items():
            log.info("Loading artifact verifier storage plugin: {0}", plugin_name)
            plugins[plugin_name] = plugin_instance
        self.artifact_verifier_ctx.load_verifiers(plugins)

    def list_volumes(self) -> Mapping[str, VolumeInfo]:
        return {name: info.to_dataclass() for name, info in self.local_config.volume.items()}

    @asynccontextmanager
    async def get_volume(self, name: str) -> AsyncIterator[AbstractVolume]:
        if name in self.volumes:
            yield self.volumes[name]
        else:
            try:
                volume_config = self.local_config.volume[name]
            except KeyError:
                raise InvalidVolumeError(name)
            volume_cls: type[AbstractVolume] = self.backends[volume_config.backend]
            volume_obj = volume_cls(
                local_config=self.local_config.model_dump(by_alias=True),
                mount_path=Path(volume_config.path),
                options=volume_config.options or {},
                etcd=self.etcd,
                event_dispatcher=self.event_dispatcher,
                event_producer=self.event_producer,
                watcher=self.watcher,
            )

            await volume_obj.init()
            self.volumes[name] = volume_obj

            yield volume_obj

    async def shutdown_volumes(self) -> None:
        for volume in self.volumes.values():
            await volume.shutdown()

    async def shutdown_manager_http_clients(self) -> None:
        """Close all manager HTTP client sessions."""
        await self.manager_client_pool.cleanup()



import enum

from pydantic import BaseModel, ConfigDict

from .address import IPAddress
from .uoid import UOID


class NodeMode(enum.StrEnum):
    ONLINE = "ONLINE"
    OFFLINE = "OFFLINE"
    MAINTENANCE = "MAINTENANCE"
    DISABLED = "DISABLED"
    UPDATE = "UPDATE"


class NodeType(enum.StrEnum):
    PD = "PD"
    NETAPP_CMODE = "NETAPP_CMODE"
    NETAPP_7MODE = "NETAPP_7MODE"
    NETAPP_CLOUD = "NETAPP_CLOUD"
    EMC_ISILON = "EMC_ISILON"
    EMC_VNX = "EMC_VNX"
    EMC_UNITY = "EMC_UNITY"
    GOOGLE_CLOUD_FILESTORE = "GOOGLE_CLOUD_FILESTORE"
    QUMULO = "QUMULO"
    ROZOFS = "ROZOFS"
    ROZOFS_HS = "ROZOFS_HS"
    SOFTNAS_CLOUD = "SOFTNAS_CLOUD"
    WINDOWS_FILE_SERVER = "WINDOWS_FILE_SERVER"
    DELL_ENAS = "DELL_ENAS"
    PURE_FB = "PURE_FB"
    VAST = "VAST"
    WEKA = "WEKA"
    NETAPP_FSX = "NETAPP_FSX"
    AMAZON_S3 = "AMAZON_S3"
    ACTIVE_SCALE_S3 = "ACTIVE_SCALE_S3"
    IBM_S3 = "IBM_S3"
    CLOUDIAN_S3 = "CLOUDIAN_S3"
    ECS_S3 = "ECS_S3"
    GENERIC_S3 = "GENERIC_S3"
    GOOGLE_S3 = "GOOGLE_S3"
    SCALITY_S3 = "SCALITY_S3"
    STORAGE_GRID_S3 = "STORAGE_GRID_S3"
    SWIFT = "SWIFT"
    AZURE = "AZURE"
    GOOGLE_CLOUD = "GOOGLE_CLOUD"
    HCP_S3 = "HCP_S3"
    ATMOS = "ATMOS"
    WASABI_S3 = "WASABI_S3"
    NETAPP_S3 = "NETAPP_S3"
    MCAFEE_AV = "MCAFEE_AV"
    CLAM_AV = "CLAM_AV"
    SNOWFLAKE = "SNOWFLAKE"
    INTERNAL_S3 = "INTERNAL_S3"
    ALCHEMI = "ALCHEMI"
    PURE_FB_S3 = "PURE_FB_S3"
    SEAGATE_LYVE_S3 = "SEAGATE_LYVE_S3"
    CARINGO_SWARM_S3 = "CARINGO_SWARM_S3"
    ISILON_S3 = "ISILON_S3"
    BACKBLAZE_S3 = "BACKBLAZE_S3"
    HAMMERSPACE_S3 = "HAMMERSPACE_S3"
    OTHER = "OTHER"
    MOVER_EXT = "MOVER_EXT"


class ProductionNodeType(enum.StrEnum):
    ANVIL = "ANVIL"
    DSX = "DSX"


class Node(BaseModel):
    model_config = ConfigDict(extra="allow")

    uoid: UOID
    name: str
    internalId: int
    mgmtIpAddress: IPAddress
    nodeMode: NodeMode
    nodeType: NodeType
    productNodeType: ProductionNodeType



"""
Base exception classes for storage proxy.
"""

from __future__ import annotations

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class StorageProxyError(BackendAIError, web.HTTPInternalServerError):
    """Base exception for all storage proxy errors."""

    error_type = "https://api.backend.ai/probs/storage/generic"
    error_title = "Storage Proxy Error"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ProcessExecutionError(BackendAIError, web.HTTPInternalServerError):
    """Raised when a storage operation execution fails."""

    error_type = "https://api.backend.ai/probs/storage/execution/failed"
    error_title = "Storage Operation Execution Failed"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.EXECUTE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ExternalStorageServiceError(BackendAIError, web.HTTPInternalServerError):
    """Raised when an external storage service operation fails."""

    error_type = "https://api.backend.ai/probs/storage/external/failed"
    error_title = "External Operation Failed"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.REQUEST,
            error_detail=ErrorDetail.UNREACHABLE,
        )


class NotImplementedAPI(BackendAIError, web.HTTPBadRequest):
    """Raised when an API is not implemented."""

    error_type = "https://api.backend.ai/probs/storage/api/not-implemented"
    error_title = "API Not Implemented"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.REQUEST,
            error_detail=ErrorDetail.NOT_IMPLEMENTED,
        )



import asyncio
import logging
from collections.abc import Mapping
from dataclasses import asdict
from pathlib import Path
from typing import Any, Final, FrozenSet, Literal, Optional, cast

import aiofiles
import aiofiles.os

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.dispatcher import EventDispatcher, EventProducer
from ai.backend.common.json import dump_json_str
from ai.backend.common.types import HardwareMetadata, QuotaConfig, QuotaScopeID
from ai.backend.logging import BraceStyleAdapter

from ...errors import (
    ExternalStorageServiceError,
    InvalidQuotaConfig,
    QuotaScopeNotFoundError,
    StorageProxyError,
)
from ...types import CapacityUsage, FSPerfMetric, QuotaUsage
from ...watcher import WatcherClient
from ..abc import CAP_FAST_FS_SIZE, CAP_FAST_SIZE, CAP_METRIC, CAP_QUOTA, CAP_VFOLDER
from ..vfs import BaseQuotaModel, BaseVolume
from .config import config_iv
from .exceptions import VASTInvalidParameterError, VASTNotFoundError, VASTUnknownError
from .vastdata_client import VASTAPIClient, VASTQuota, VASTQuotaID

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


VAST_QUOTA_ID_FILE_NAME: Final = ".vast-quota-id"


class VASTQuotaModel(BaseQuotaModel):
    def __init__(
        self,
        mount_path: Path,
        api_client: VASTAPIClient,
    ) -> None:
        super().__init__(mount_path)
        self.api_client = api_client

    async def _get_vast_quota_id(self, quota_scope_id: QuotaScopeID) -> VASTQuotaID | None:
        qs_path = self.mangle_qspath(quota_scope_id)

        def _read():
            try:
                with open(qs_path / VAST_QUOTA_ID_FILE_NAME, "r") as f:
                    return VASTQuotaID(f.read())
            except FileNotFoundError:
                return None

        return await asyncio.get_running_loop().run_in_executor(None, _read)

    async def _set_vast_quota_id(
        self, quota_scope_id: QuotaScopeID, vast_quota_id: VASTQuotaID
    ) -> None:
        qs_path = self.mangle_qspath(quota_scope_id)

        def _write():
            qs_path.mkdir(parents=True, exist_ok=True)
            with open(qs_path / VAST_QUOTA_ID_FILE_NAME, "w") as f:
                f.write(str(vast_quota_id))

        await asyncio.get_running_loop().run_in_executor(None, _write)

    async def _rm_vast_quota_id(self, quota_scope_id: QuotaScopeID) -> None:
        qs_path = self.mangle_qspath(quota_scope_id)

        try:
            await aiofiles.os.remove(qs_path / VAST_QUOTA_ID_FILE_NAME)
        except FileNotFoundError:
            log.warning(f"vast quota id file not found (qid: {quota_scope_id}). skip")

    async def _modify_quota_scope(
        self,
        vast_quota_id: VASTQuotaID,
        config: QuotaConfig,
    ) -> VASTQuota:
        try:
            return await self.api_client.modify_quota(
                vast_quota_id,
                soft_limit=config.limit_bytes,
                hard_limit=config.limit_bytes,
            )
        except VASTInvalidParameterError:
            raise InvalidQuotaConfig
        except VASTUnknownError as e:
            raise ExternalStorageServiceError(
                f"VAST API unknown error during quota modification: {e}"
            )

    async def create_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        options: Optional[QuotaConfig] = None,
        extra_args: Optional[dict[str, Any]] = None,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)

        async def _set_quota(_options: QuotaConfig) -> VASTQuota:
            try:
                quota = await self.api_client.set_quota(
                    qspath,
                    soft_limit=_options.limit_bytes,
                    hard_limit=_options.limit_bytes,
                )
            except VASTInvalidParameterError as e:
                # Check if quota has already been set to the quota scope path
                quota_name = str(qspath)
                quotas = await self.api_client.list_quotas(quota_name)
                existing_quota: Optional[VASTQuota] = None
                for q in quotas:
                    if q.name == quota_name:
                        existing_quota = q
                        break
                else:
                    log.error(
                        "Got invalid parameter error but no quota exists with given quota name"
                        f" ({quota_name}). Raise error (orig:{str(e)})"
                    )
                    raise InvalidQuotaConfig(f"No existing quota found with name {quota_name}")
                await self._set_vast_quota_id(quota_scope_id, existing_quota.id)
                await self.api_client.modify_quota(
                    existing_quota.id,
                    soft_limit=_options.limit_bytes,
                    hard_limit=_options.limit_bytes,
                )
                return existing_quota
            except VASTUnknownError as e:
                raise ExternalStorageServiceError(
                    f"VAST API unknown error during quota creation: {e}"
                )
            return quota

        try:
            await aiofiles.os.makedirs(qspath)
        except FileExistsError:
            if options is None:
                return
            vast_quota_id = await self._get_vast_quota_id(quota_scope_id)
            if vast_quota_id is not None:
                existing_quota = await self.api_client.get_quota(vast_quota_id)
                if existing_quota is not None:
                    quota = await self._modify_quota_scope(vast_quota_id, options)
                else:
                    quota = await _set_quota(options)
            else:
                quota = await _set_quota(options)
            await self._set_vast_quota_id(quota_scope_id, quota.id)
        else:
            if options is None:
                return
            quota = await _set_quota(options)
            await self._set_vast_quota_id(quota_scope_id, quota.id)

    async def update_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        config: QuotaConfig,
    ) -> None:
        vast_quota_id = await self._get_vast_quota_id(quota_scope_id)
        if vast_quota_id is None:
            raise QuotaScopeNotFoundError
        await self._modify_quota_scope(vast_quota_id, config)

    async def describe_quota_scope(self, quota_scope_id: QuotaScopeID) -> Optional[QuotaUsage]:
        if (vast_quota_id := await self._get_vast_quota_id(quota_scope_id)) is None:
            return None
        if (quota := await self.api_client.get_quota(vast_quota_id)) is None:
            return None
        if quota.used_capacity < 0 or quota.hard_limit < 0:
            log.warning(
                "Data from VAST API negative values in used_bytes({}) or limit_bytes({}) for quota scope {}: response from VAST API = {}",
                quota.used_capacity,
                quota.hard_limit,
                quota_scope_id,
                quota,
            )
        return QuotaUsage(
            used_bytes=quota.used_capacity,
            limit_bytes=quota.hard_limit,
        )

    async def unset_quota(self, quota_scope_id: QuotaScopeID) -> None:
        vast_quota_id = await self._get_vast_quota_id(quota_scope_id)
        if vast_quota_id is None:
            raise QuotaScopeNotFoundError
        try:
            await self.api_client.remove_quota(vast_quota_id)
        except VASTNotFoundError:
            raise QuotaScopeNotFoundError
        await self._rm_vast_quota_id(quota_scope_id)

    async def delete_quota_scope(self, quota_scope_id: QuotaScopeID) -> None:
        await self.unset_quota(quota_scope_id)
        qspath = self.mangle_qspath(quota_scope_id)
        await aiofiles.os.rmdir(qspath)


class VASTVolume(BaseVolume):
    api_client: VASTAPIClient

    name = "vast"

    def __init__(
        self,
        local_config: Mapping[str, Any],
        mount_path: Path,
        *,
        etcd: AsyncEtcd,
        event_dispatcher: EventDispatcher,
        event_producer: EventProducer,
        watcher: Optional[WatcherClient] = None,
        options: Optional[Mapping[str, Any]] = None,
    ) -> None:
        super().__init__(
            local_config,
            mount_path,
            etcd=etcd,
            options=options,
            event_dispatcher=event_dispatcher,
            event_producer=event_producer,
            watcher=watcher,
        )
        self.config = cast(Mapping[str, Any], config_iv.check(self.config))
        ssl_verify = self.config.get("vast_verify_ssl", False)
        self.api_client = VASTAPIClient(
            self.config["vast_endpoint"],
            self.config["vast_username"],
            self.config["vast_password"],
            storage_base_dir=self.config["vast_storage_base_dir"],
            api_version=self.config["vast_api_version"],
            ssl=ssl_verify,
            force_login=self.config["vast_force_login"],
            cluster_info_cache_ttl=self.config["vast_cluster_info_cache_ttl"],
        )

    async def shutdown(self) -> None:
        pass

    async def create_quota_model(self) -> VASTQuotaModel:
        return VASTQuotaModel(self.mount_path, self.api_client)

    async def get_capabilities(self) -> FrozenSet[str]:
        return frozenset([CAP_VFOLDER, CAP_METRIC, CAP_QUOTA, CAP_FAST_FS_SIZE, CAP_FAST_SIZE])

    async def get_hwinfo(self) -> HardwareMetadata:
        cluster_id: int = self.config["vast_cluster_id"]
        try:
            clsuter_info = await self.api_client.get_cluster_info(cluster_id)
        except VASTUnknownError:
            return {
                "status": "unavailable",
                "status_info": None,
                "metadata": {},
            }
        if clsuter_info is None:
            raise StorageProxyError(f"vast cluster not found. (id: {cluster_id})")
        healthy_status: Literal["healthy", "degraded", "unavailable"] = "unavailable"
        match clsuter_info.state.lower():
            case "online" | "healthy":
                healthy_status = "healthy"
            case "init":
                healthy_status = "degraded"
            case "unknown":
                healthy_status = "unavailable"
        quotas = await self.api_client.list_quotas()
        return {
            "status": healthy_status,
            "status_info": clsuter_info.state,
            "metadata": {
                "quota": dump_json_str([asdict(q) for q in quotas]),
                "cluster_info": dump_json_str(asdict(clsuter_info)),
            },
        }

    async def get_performance_metric(self) -> FSPerfMetric:
        cluster_id: int = self.config["vast_cluster_id"]
        try:
            clsuter_info = await self.api_client.get_cluster_info(cluster_id)
        except VASTUnknownError:
            return FSPerfMetric(
                iops_read=-1,
                iops_write=-1,
                io_bytes_read=-1,
                io_bytes_write=-1,
                io_usec_read=-1,
                io_usec_write=-1,
            )
        if clsuter_info is None:
            raise StorageProxyError(f"vast cluster not found. (id: {cluster_id})")
        return FSPerfMetric(
            iops_read=clsuter_info.rd_iops,
            iops_write=clsuter_info.wr_iops,
            io_usec_read=clsuter_info.rd_latency,
            io_usec_write=clsuter_info.wr_latency,
            io_bytes_read=clsuter_info.rd_bw,
            io_bytes_write=clsuter_info.wr_bw,
        )

    async def get_fs_usage(self) -> CapacityUsage:
        cluster_id: int = self.config["vast_cluster_id"]
        try:
            clsuter_info = await self.api_client.get_cluster_info(cluster_id)
        except VASTUnknownError:
            return CapacityUsage(
                used_bytes=-1,
                capacity_bytes=-1,
            )
        if clsuter_info is None:
            raise StorageProxyError(f"vast cluster not found. (id: {cluster_id})")
        return CapacityUsage(
            used_bytes=clsuter_info.physical_space_in_use,
            capacity_bytes=clsuter_info.physical_space,
        )



from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class HammerspaceError(BackendAIError):
    """Base error for Hammerspace-related errors."""


class ConfigurationError(HammerspaceError):
    error_type = "https://api.backend.ai/probs/hammerspace-invalid-configuration"
    error_title = "Hammerspace configuration is invalid."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.SETUP,
            error_detail=ErrorDetail.NOT_READY,
        )


class AuthenticationError(HammerspaceError, web.HTTPUnauthorized):
    error_type = "https://api.backend.ai/probs/hammerspace-authentication-failure"
    error_title = "Hammerspace API authentication failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.UNAUTHORIZED,
        )


class ObjectiveNotFound(HammerspaceError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/hammerspace-objective-not-found"
    error_title = "Hammerspace has no such objective."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ShareNotFound(HammerspaceError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/hammerspace-share-not-found"
    error_title = "Hammerspace has no such share."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class VolumeNotFound(HammerspaceError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/hammerspace-volume-not-found"
    error_title = "Hammerspace has no such volume."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )



"""
Common/shared exceptions for storage proxy.
"""

from __future__ import annotations

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class InvalidAPIParameters(BackendAIError, web.HTTPBadRequest):
    """Raised when API parameters are invalid."""

    error_type = "https://api.backend.ai/probs/storage/invalid-api-params"
    error_title = "Invalid API parameters"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.REQUEST,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class StorageNotFoundError(BackendAIError, web.HTTPNotFound):
    """Raised when a storage config is not found."""

    error_type = "https://api.backend.ai/probs/storage/object-not-found"
    error_title = "Storage Config Not Found"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class StorageTypeInvalidError(BackendAIError, web.HTTPBadRequest):
    """Raised when a storage type is invalid."""

    error_type = "https://api.backend.ai/probs/storage/object-type-invalid"
    error_title = "Storage Config Invalid Type"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class StorageTransferError(BackendAIError, web.HTTPInternalServerError):
    """Raised when a storage transfer operation fails."""

    error_type = "https://api.backend.ai/probs/storage/transfer/failed"
    error_title = "Storage Transfer Failed"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class StorageStepRequiredStepNotProvided(BackendAIError, web.HTTPBadRequest):
    """Raised when a required storage step mapping is not provided."""

    error_type = "https://api.backend.ai/probs/storage/step-mapping-not-provided"
    error_title = "Storage Step Mapping Not Provided"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.SETUP,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


# Path validation exceptions


class InvalidPathError(BackendAIError, web.HTTPBadRequest):
    """Raised when a path is invalid (empty, contains invalid characters, or attempts traversal)."""

    error_type = "https://api.backend.ai/probs/storage/path/invalid"
    error_title = "Invalid Path"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class InvalidSocketPathError(BackendAIError, web.HTTPBadRequest):
    """Raised when a socket path is invalid or not provided."""

    error_type = "https://api.backend.ai/probs/storage/socket-path/invalid"
    error_title = "Invalid Socket Path"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class InvalidConfigurationSourceError(BackendAIError, web.HTTPInternalServerError):
    """Raised when a configuration source is invalid."""

    error_type = "https://api.backend.ai/probs/storage/config/invalid-source"
    error_title = "Invalid Configuration Source"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.SETUP,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class InvalidDataLengthError(BackendAIError, web.HTTPBadRequest):
    """Raised when data has an invalid length."""

    error_type = "https://api.backend.ai/probs/storage/data/invalid-length"
    error_title = "Invalid Data Length"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.REQUEST,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class ServiceNotInitializedError(BackendAIError, web.HTTPInternalServerError):
    """Raised when a service is not properly initialized."""

    error_type = "https://api.backend.ai/probs/storage/service/not-initialized"
    error_title = "Service Not Initialized"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.SETUP,
            error_detail=ErrorDetail.NOT_READY,
        )



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from ai.backend.common.dependencies import DependencyComposer, DependencyStack
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.storage.config.unified import StorageProxyUnifiedConfig

from .etcd import EtcdProvider
from .redis import RedisProvider, StorageProxyValkeyClients


@dataclass
class InfrastructureComposerInput:
    """Input for Infrastructure composer."""

    local_config: StorageProxyUnifiedConfig


@dataclass
class InfrastructureResources:
    """All infrastructure resources for storage proxy."""

    etcd: AsyncEtcd
    valkey: StorageProxyValkeyClients


class InfrastructureComposer(
    DependencyComposer[InfrastructureComposerInput, InfrastructureResources]
):
    """Composer for infrastructure layer dependencies."""

    @property
    def stage_name(self) -> str:
        return "infrastructure"

    @asynccontextmanager
    async def compose(
        self,
        stack: DependencyStack,
        setup_input: InfrastructureComposerInput,
    ) -> AsyncIterator[InfrastructureResources]:
        """Compose all infrastructure dependencies."""
        local_config = setup_input.local_config

        # Setup infrastructure in dependency order
        etcd = await stack.enter_dependency(EtcdProvider(), local_config)
        valkey = await stack.enter_dependency(RedisProvider(), etcd)

        yield InfrastructureResources(
            etcd=etcd,
            valkey=valkey,
        )



from __future__ import annotations

import asyncio
import functools
import grp
import logging
import multiprocessing
import os
import pwd
import signal
import ssl
import sys
import traceback
from contextlib import AsyncExitStack, asynccontextmanager
from pathlib import Path
from pprint import pformat, pprint
from typing import TYPE_CHECKING, Any, AsyncGenerator, AsyncIterator, Sequence

import aiohttp_cors
import aiomonitor
import aiotools
import click
from aiohttp import web
from aiohttp.typedefs import Middleware
from setproctitle import setproctitle

from ai.backend.common.bgtask.bgtask import BackgroundTaskManager
from ai.backend.common.clients.valkey_client.valkey_artifact.client import (
    ValkeyArtifactDownloadTrackingClient,
)
from ai.backend.common.clients.valkey_client.valkey_bgtask.client import ValkeyBgtaskClient
from ai.backend.common.config import (
    ConfigurationError,
)
from ai.backend.common.configs.redis import RedisConfig
from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.common.defs import (
    NOOP_STORAGE_VOLUME_NAME,
    REDIS_BGTASK_DB,
    REDIS_STATISTICS_DB,
    REDIS_STREAM_DB,
    RedisRole,
)
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.dispatcher import EventDispatcher, EventProducer
from ai.backend.common.health_checker.checkers.etcd import EtcdHealthChecker
from ai.backend.common.health_checker.checkers.valkey import ValkeyHealthChecker
from ai.backend.common.health_checker.probe import HealthProbe, HealthProbeOptions
from ai.backend.common.health_checker.types import ComponentId
from ai.backend.common.message_queue.hiredis_queue import HiRedisQueue
from ai.backend.common.message_queue.queue import AbstractMessageQueue
from ai.backend.common.message_queue.redis_queue import RedisMQArgs, RedisQueue
from ai.backend.common.metrics.metric import CommonMetricRegistry
from ai.backend.common.metrics.profiler import Profiler, PyroscopeArgs
from ai.backend.common.msgpack import DEFAULT_PACK_OPTS, DEFAULT_UNPACK_OPTS
from ai.backend.common.plugin import BasePluginContext
from ai.backend.common.service_discovery.etcd_discovery.service_discovery import (
    ETCDServiceDiscovery,
    ETCDServiceDiscoveryArgs,
)
from ai.backend.common.service_discovery.redis_discovery.service_discovery import (
    RedisServiceDiscovery,
    RedisServiceDiscoveryArgs,
)
from ai.backend.common.service_discovery.service_discovery import (
    ServiceDiscovery,
    ServiceDiscoveryLoop,
    ServiceEndpoint,
    ServiceMetadata,
)
from ai.backend.common.types import (
    AGENTID_STORAGE,
    RedisProfileTarget,
    ServiceDiscoveryType,
    safe_print_redis_config,
)
from ai.backend.common.types import HostPortPair as CommonHostPortPair
from ai.backend.common.utils import env_info
from ai.backend.logging import BraceStyleAdapter, Logger, LogLevel
from ai.backend.logging.otel import OpenTelemetrySpec
from ai.backend.storage.context_types import ArtifactVerifierContext

from . import __version__ as VERSION
from .client.manager import ManagerHTTPClientPool
from .config.loaders import load_local_config, make_etcd
from .config.unified import (
    EventLoopType,
    LegacyReservoirConfig,
    ReservoirConfig,
    StorageProxyUnifiedConfig,
)
from .errors import InvalidConfigurationSourceError, InvalidSocketPathError
from .watcher import WatcherClient

if TYPE_CHECKING:
    from .context import RootContext
    from .volumes.pool import VolumePool

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


def _is_root() -> bool:
    return os.geteuid() == 0


@aiotools.server_context
async def server_main_logwrapper(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    _args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    setproctitle(f"backend.ai: storage-proxy worker-{pidx}")
    local_config: StorageProxyUnifiedConfig = _args[0]
    log_endpoint = _args[1]
    logger = Logger(
        local_config.logging,
        is_master=False,
        log_endpoint=log_endpoint,
        msgpack_options={
            "pack_opts": DEFAULT_PACK_OPTS,
            "unpack_opts": DEFAULT_UNPACK_OPTS,
        },
    )
    try:
        with logger:
            async with server_main(loop, pidx, _args):
                yield
    except Exception:
        traceback.print_exc(file=sys.stderr)


@asynccontextmanager
async def aiomonitor_ctx(
    local_config: StorageProxyUnifiedConfig,
    pidx: int,
) -> AsyncGenerator[aiomonitor.Monitor]:
    # Port is set by config where the defaults are:
    # termui_port = 38300 + pidx
    # webui_port = 39300 + pidx
    loop = asyncio.get_running_loop()
    m = aiomonitor.Monitor(
        loop,
        termui_port=local_config.storage_proxy.aiomonitor_termui_port + pidx,
        webui_port=local_config.storage_proxy.aiomonitor_webui_port + pidx,
        console_enabled=False,
        hook_task_factory=local_config.debug.enhanced_aiomonitor_task_info,
    )
    Profiler(
        pyroscope_args=PyroscopeArgs(
            enabled=local_config.pyroscope.enabled,
            application_name=local_config.pyroscope.app_name,
            server_address=local_config.pyroscope.server_addr,
            sample_rate=local_config.pyroscope.sample_rate,
        )
    )
    m.prompt = f"monitor (storage-proxy[{pidx}@{os.getpid()}]) >>> "
    m.console_locals["local_config"] = local_config
    aiomon_started = False
    try:
        m.start()
        aiomon_started = True
    except Exception as e:
        log.warning("aiomonitor could not start but skipping this error to continue", exc_info=e)
    try:
        yield m
    finally:
        if aiomon_started:
            m.close()


@asynccontextmanager
async def etcd_ctx(local_config: StorageProxyUnifiedConfig) -> AsyncGenerator[AsyncEtcd]:
    etcd = make_etcd(local_config)
    try:
        yield etcd
    finally:
        await etcd.close()


@asynccontextmanager
async def redis_ctx(etcd: AsyncEtcd, pidx: int) -> AsyncGenerator[RedisConfig]:
    raw_redis_config = await etcd.get_prefix("config/redis")
    redis_config = RedisConfig.model_validate(raw_redis_config)
    log.info(
        "PID: {0} - configured redis_config: {1}",
        pidx,
        safe_print_redis_config(redis_config),
    )
    yield redis_config


@asynccontextmanager
async def bgtask_ctx(
    local_config: StorageProxyUnifiedConfig,
    redis_config: RedisConfig,
    event_producer: EventProducer,
    volume_pool: VolumePool,
) -> AsyncGenerator[BackgroundTaskManager]:
    from .bgtask.registry import BgtaskHandlerRegistryCreator

    redis_profile_target = redis_config.to_redis_profile_target()
    valkey_client = await ValkeyBgtaskClient.create(
        redis_profile_target.profile_target(RedisRole.BGTASK).to_valkey_target(),
        human_readable_name="storage_bgtask",
        db_id=REDIS_BGTASK_DB,
    )
    bgtask_registry_creator = BgtaskHandlerRegistryCreator(volume_pool, event_producer)
    registry = bgtask_registry_creator.create()
    bgtask_mgr = BackgroundTaskManager(
        event_producer=event_producer,
        task_registry=registry,
        valkey_client=valkey_client,
        server_id=local_config.storage_proxy.node_id,
    )

    try:
        yield bgtask_mgr
    finally:
        await valkey_client.close()


async def _make_message_queue(
    local_config: StorageProxyUnifiedConfig,
    redis_profile_target: RedisProfileTarget,
) -> AbstractMessageQueue:
    from .context import EVENT_DISPATCHER_CONSUMER_GROUP

    stream_redis_target = redis_profile_target.profile_target(RedisRole.STREAM)
    node_id = local_config.storage_proxy.node_id
    args = RedisMQArgs(
        anycast_stream_key="events",
        broadcast_channel="events_all",
        consume_stream_keys=None,
        subscribe_channels={
            "events_all",
        },
        group_name=EVENT_DISPATCHER_CONSUMER_GROUP,
        node_id=node_id,
        db=REDIS_STREAM_DB,
    )
    if local_config.storage_proxy.use_experimental_redis_event_dispatcher:
        return HiRedisQueue(
            stream_redis_target,
            args,
        )
    return await RedisQueue.create(
        redis_profile_target.profile_target(RedisRole.STREAM),
        args,
    )


@asynccontextmanager
async def event_ctx(
    local_config: StorageProxyUnifiedConfig,
    redis_config: RedisConfig,
    pidx: int,
    metric_registry: CommonMetricRegistry,
) -> AsyncGenerator[tuple[EventDispatcher, EventProducer]]:
    redis_profile_target = redis_config.to_redis_profile_target()
    mq = await _make_message_queue(
        local_config,
        redis_profile_target,
    )
    event_producer = EventProducer(
        mq,
        source=AGENTID_STORAGE,
        log_events=local_config.debug.log_events,
    )
    log.info(
        "PID: {0} - Event producer created. (redis_config: {1})",
        pidx,
        safe_print_redis_config(redis_config),
    )
    event_dispatcher = EventDispatcher(
        mq,
        log_events=local_config.debug.log_events,
        event_observer=metric_registry.event,
    )
    log.info(
        "PID: {0} - Event dispatcher created. (redis_config: {1})",
        pidx,
        safe_print_redis_config(redis_config),
    )
    await event_dispatcher.start()
    try:
        yield event_dispatcher, event_producer
    finally:
        await event_producer.close()
        await event_dispatcher.close()


@asynccontextmanager
async def watcher_ctx(
    local_config: StorageProxyUnifiedConfig,
    pidx: int,
) -> AsyncGenerator[WatcherClient | None]:
    if local_config.storage_proxy.use_watcher:
        if not _is_root():
            raise InvalidConfigurationSourceError(
                "Storage proxy must be run as root if watcher is enabled. Else, set"
                " `use-watcher` to false in your local config file."
            )
        insock_path: str | None = local_config.storage_proxy.watcher_insock_path_prefix
        outsock_path: str | None = local_config.storage_proxy.watcher_outsock_path_prefix
        if insock_path is None or outsock_path is None:
            raise InvalidSocketPathError(
                "Socket path must be not null. Please set valid socket path to"
                " `watcher-insock-path-prefix` and `watcher-outsock-path-prefix` in your local"
                " config file."
            )
        watcher_client = WatcherClient(pidx, insock_path, outsock_path)
        await watcher_client.init()
    else:
        watcher_client = None
    try:
        yield watcher_client
    finally:
        if watcher_client is not None:
            await watcher_client.close()


@asynccontextmanager
async def volume_ctx(
    local_config: StorageProxyUnifiedConfig,
    etcd: AsyncEtcd,
    event_dispatcher: EventDispatcher,
    event_producer: EventProducer,
) -> AsyncGenerator[VolumePool]:
    from .volumes.pool import VolumePool

    volume_pool = await VolumePool.create(
        local_config=local_config,
        etcd=etcd,
        event_dispatcher=event_dispatcher,
        event_producer=event_producer,
    )
    try:
        yield volume_pool
    finally:
        await volume_pool.shutdown()


@asynccontextmanager
async def api_ctx(
    local_config: StorageProxyUnifiedConfig,
    etcd: AsyncEtcd,
    root_ctx: RootContext,
) -> AsyncGenerator[tuple[web.Application, web.Application, web.Application]]:
    from .api.client import init_client_app
    from .api.manager import init_internal_app, init_manager_app
    from .plugin import (
        StorageClientWebappPluginContext,
        StorageManagerWebappPluginContext,
        StoragePluginContext,
    )

    @asynccontextmanager
    async def _init_storage_plugin() -> AsyncGenerator[StoragePluginContext]:
        plugin_ctx = StoragePluginContext(etcd, local_config.model_dump())
        await plugin_ctx.init()
        for plugin_name, plugin_instance in plugin_ctx.plugins.items():
            log.info("Loading storage plugin: {0}", plugin_name)
            volume_cls = plugin_instance.get_volume_class()
            root_ctx.backends[plugin_name] = volume_cls
        try:
            yield plugin_ctx
        finally:
            await plugin_ctx.cleanup()

    @asynccontextmanager
    async def _init_storage_webapp_plugin(
        plugin_ctx: BasePluginContext, root_app: web.Application
    ) -> AsyncGenerator[BasePluginContext]:
        pid = os.getpid()
        await plugin_ctx.init()
        for plugin_name, plugin_instance in plugin_ctx.plugins.items():
            if pid == 0:
                log.info("Loading storage webapp plugin: {0}", plugin_name)
            subapp, global_middlewares = await plugin_instance.create_app(root_ctx.cors_options)
            _init_subapp(plugin_name, root_app, subapp, global_middlewares)
        try:
            yield plugin_ctx
        finally:
            await plugin_ctx.cleanup()

    @asynccontextmanager
    async def client_api_ctx() -> AsyncGenerator[web.Application]:
        client_ssl_ctx = None
        if local_config.api.client.ssl_enabled:
            client_ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            client_ssl_ctx.load_cert_chain(
                str(local_config.api.client.ssl_cert),
                str(local_config.api.client.ssl_privkey),
            )
        client_api_app = await init_client_app(root_ctx)
        client_api_runner = web.AppRunner(client_api_app)
        await client_api_runner.setup()
        client_service_addr = local_config.api.client.service_addr
        client_api_site = web.TCPSite(
            client_api_runner,
            str(client_service_addr.host),
            client_service_addr.port,
            backlog=1024,
            reuse_port=True,
            ssl_context=client_ssl_ctx,
        )
        await client_api_site.start()
        try:
            yield client_api_app
        finally:
            await client_api_runner.cleanup()

    @asynccontextmanager
    async def manager_api_ctx() -> AsyncGenerator[web.Application]:
        manager_ssl_ctx = None
        if local_config.api.manager.ssl_enabled:
            manager_ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            manager_ssl_ctx.load_cert_chain(
                str(local_config.api.manager.ssl_cert),
                str(local_config.api.manager.ssl_privkey),
            )
        manager_api_app = await init_manager_app(root_ctx)
        manager_api_runner = web.AppRunner(manager_api_app)
        await manager_api_runner.setup()
        manager_service_addr_config = local_config.api.manager.service_addr
        manager_service_addr = CommonHostPortPair(
            host=manager_service_addr_config.host,
            port=manager_service_addr_config.port,
        )
        manager_api_site = web.TCPSite(
            manager_api_runner,
            str(manager_service_addr.host),
            manager_service_addr.port,
            backlog=1024,
            reuse_port=True,
            ssl_context=manager_ssl_ctx,
        )
        await manager_api_site.start()
        try:
            yield manager_api_app
        finally:
            await manager_api_runner.cleanup()

    @asynccontextmanager
    async def internal_api_ctx() -> AsyncGenerator[web.Application]:
        internal_api_app = init_internal_app(root_ctx)
        internal_api_runner = web.AppRunner(internal_api_app)
        await internal_api_runner.setup()
        internal_addr = local_config.api.manager.internal_addr
        internal_api_site = web.TCPSite(
            internal_api_runner,
            str(internal_addr.host),
            internal_addr.port,
            backlog=1024,
            reuse_port=True,
        )
        await internal_api_site.start()
        try:
            yield internal_api_app
        finally:
            await internal_api_runner.cleanup()

    async with AsyncExitStack() as api_init_stack:
        await api_init_stack.enter_async_context(_init_storage_plugin())
        client_api_app = await api_init_stack.enter_async_context(client_api_ctx())
        manager_api_app = await api_init_stack.enter_async_context(manager_api_ctx())
        internal_api_app = await api_init_stack.enter_async_context(internal_api_ctx())
        await api_init_stack.enter_async_context(
            _init_storage_webapp_plugin(
                StorageClientWebappPluginContext(etcd, local_config.model_dump()),
                client_api_app,
            )
        )
        await api_init_stack.enter_async_context(
            _init_storage_webapp_plugin(
                StorageManagerWebappPluginContext(etcd, local_config.model_dump()),
                manager_api_app,
            )
        )
        try:
            yield client_api_app, manager_api_app, internal_api_app
        finally:
            # volume instances are lazily initialized upon their first usage by the API layers.
            await root_ctx.shutdown_volumes()
            await root_ctx.shutdown_manager_http_clients()


@asynccontextmanager
async def service_discovery_ctx(
    local_config: StorageProxyUnifiedConfig,
    etcd: AsyncEtcd,
    redis_config: RedisConfig,
) -> AsyncGenerator[None]:
    announce_addr_config = local_config.api.manager.announce_addr
    announce_addr = CommonHostPortPair(
        host=announce_addr_config.host,
        port=announce_addr_config.port,
    )
    announce_internal_addr_config = local_config.api.manager.announce_internal_addr
    announce_internal_addr = CommonHostPortPair(
        host=announce_internal_addr_config.host,
        port=announce_internal_addr_config.port,
    )

    sd_type = local_config.service_discovery.type

    service_discovery: ServiceDiscovery
    match sd_type:
        case ServiceDiscoveryType.ETCD:
            service_discovery = ETCDServiceDiscovery(ETCDServiceDiscoveryArgs(etcd))
        case ServiceDiscoveryType.REDIS:
            valkey_profile_target = redis_config.to_valkey_profile_target()
            live_valkey_target = valkey_profile_target.profile_target(RedisRole.LIVE)
            service_discovery = await RedisServiceDiscovery.create(
                args=RedisServiceDiscoveryArgs(valkey_target=live_valkey_target)
            )

    sd_loop = ServiceDiscoveryLoop(
        sd_type,
        service_discovery,
        ServiceMetadata(
            display_name=f"storage-{local_config.storage_proxy.node_id}",
            service_group="storage-proxy",
            version=VERSION,
            endpoint=ServiceEndpoint(
                address=str(announce_addr),
                port=announce_addr.port,
                protocol="http",
                prometheus_address=str(announce_internal_addr),
            ),
        ),
    )
    if local_config.otel.enabled:
        meta = sd_loop.metadata
        otel_spec = OpenTelemetrySpec(
            service_id=meta.id,
            service_name=meta.service_group,
            service_version=meta.version,
            log_level=local_config.otel.log_level,
            endpoint=local_config.otel.endpoint,
        )
        BraceStyleAdapter.apply_otel(otel_spec)
    try:
        yield
    finally:
        sd_loop.close()


async def _on_prepare(request: web.Request, response: web.StreamResponse) -> None:
    response.headers["Server"] = "BackendAI"


def _init_subapp(
    pkg_name: str,
    root_app: web.Application,
    subapp: web.Application,
    global_middlewares: list[Middleware],
) -> None:
    subapp.on_response_prepare.append(_on_prepare)

    async def _set_root_ctx(subapp: web.Application):
        # Allow subapp's access to the root app properties.
        # These are the public APIs exposed to plugins as well.
        subapp["ctx"] = root_app["ctx"]

    # We must copy the public interface prior to all user-defined startup signal handlers.
    subapp.on_startup.insert(0, _set_root_ctx)
    if "prefix" not in subapp:
        subapp["prefix"] = pkg_name.split(".")[-1].replace("_", "-")
    prefix = subapp["prefix"]
    root_app.add_subapp("/" + prefix, subapp)
    root_app.middlewares.extend(global_middlewares)


@asynccontextmanager
async def server_main(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    _args: Sequence[Any],
) -> AsyncIterator[None]:
    from .context import DEFAULT_BACKENDS, RootContext
    from .migration import check_latest
    from .storages.storage_pool import StoragePool
    from .volumes.noop import init_noop_volume

    local_config: StorageProxyUnifiedConfig = _args[0]
    loop.set_debug(local_config.debug.asyncio)

    storage_init_stack = AsyncExitStack()
    await storage_init_stack.__aenter__()
    try:
        metric_registry = CommonMetricRegistry()
        monitor = await storage_init_stack.enter_async_context(aiomonitor_ctx(local_config, pidx))
        etcd = await storage_init_stack.enter_async_context(etcd_ctx(local_config))
        redis_config = await storage_init_stack.enter_async_context(redis_ctx(etcd, pidx))
        event_dispatcher, event_producer = await storage_init_stack.enter_async_context(
            event_ctx(local_config, redis_config, pidx, metric_registry)
        )

        # Create StoragePool with both object storage and VFS storage
        volume_pool = await storage_init_stack.enter_async_context(
            volume_ctx(local_config, etcd, event_dispatcher, event_producer)
        )
        storage_pool = StoragePool.from_config(local_config)

        # Clean up temporary storages only on the first process
        if pidx == 0:
            storage_pool.cleanup_temporary_storages()

        bgtask_mgr = await storage_init_stack.enter_async_context(
            bgtask_ctx(local_config, redis_config, event_producer, volume_pool)
        )
        watcher_client = await storage_init_stack.enter_async_context(
            watcher_ctx(local_config, pidx)
        )

        # Create ValkeyArtifactDownloadTrackingClient
        valkey_target = redis_config.to_valkey_target()
        valkey_artifact_client = await ValkeyArtifactDownloadTrackingClient.create(
            valkey_target=valkey_target,
            db_id=REDIS_STATISTICS_DB,
            human_readable_name=f"storage-proxy-artifact-download-tracker-{pidx}",
        )
        storage_init_stack.push_async_callback(valkey_artifact_client.close)

        # Initialize health probe
        health_probe = HealthProbe(options=HealthProbeOptions(check_interval=60))
        await health_probe.register(EtcdHealthChecker(etcd=etcd))
        await health_probe.register(
            ValkeyHealthChecker(
                clients={
                    ComponentId("bgtask"): bgtask_mgr._valkey_client,
                    ComponentId("artifact"): valkey_artifact_client,
                }
            )
        )
        await health_probe.start()
        storage_init_stack.push_async_callback(health_probe.stop)

        # Build reservoir registry configs for ManagerHTTPClientPool
        reservoir_registry_configs: dict[str, ReservoirConfig] = {
            name: r.reservoir
            for name, r in local_config.artifact_registries.items()
            if r.registry_type == ArtifactRegistryType.RESERVOIR and r.reservoir is not None
        }
        for legacy_registry in local_config.registries:
            if isinstance(legacy_registry.config, LegacyReservoirConfig):
                reservoir_registry_configs[legacy_registry.name] = legacy_registry.config

        manager_client_pool = ManagerHTTPClientPool(
            reservoir_registry_configs,
            local_config.reservoir_client,
        )

        root_ctx = RootContext(
            pid=os.getpid(),
            pidx=pidx,
            node_id=local_config.storage_proxy.node_id,
            local_config=local_config,
            etcd=etcd,
            volume_pool=volume_pool,
            storage_pool=storage_pool,
            background_task_manager=bgtask_mgr,
            event_producer=event_producer,
            event_dispatcher=event_dispatcher,
            watcher=watcher_client,
            metric_registry=metric_registry,
            cors_options={
                "*": aiohttp_cors.ResourceOptions(
                    allow_credentials=False, expose_headers="*", allow_headers="*"
                ),
            },
            manager_client_pool=manager_client_pool,
            valkey_artifact_client=valkey_artifact_client,
            health_probe=health_probe,
            backends={**DEFAULT_BACKENDS},
            volumes={
                NOOP_STORAGE_VOLUME_NAME: init_noop_volume(etcd, event_dispatcher, event_producer)
            },
            artifact_verifier_ctx=ArtifactVerifierContext(),
        )
        await root_ctx.init_storage_artifact_verifier_plugin()
        if pidx == 0:
            await check_latest(root_ctx)

        (
            client_api_app,
            manager_api_app,
            internal_api_app,
        ) = await storage_init_stack.enter_async_context(api_ctx(local_config, etcd, root_ctx))
        monitor.console_locals["root_ctx"] = root_ctx
        monitor.console_locals["client_api_app"] = client_api_app
        monitor.console_locals["manager_api_app"] = manager_api_app
        monitor.console_locals["internal_api_app"] = internal_api_app

        await storage_init_stack.enter_async_context(
            service_discovery_ctx(local_config, etcd, redis_config)
        )

        if _is_root():
            uid = local_config.storage_proxy.user
            gid = local_config.storage_proxy.group
            if uid is not None and gid is not None:
                os.setgroups(
                    [g.gr_gid for g in grp.getgrall() if pwd.getpwuid(uid).pw_name in g.gr_mem],
                )
                os.setgid(gid)
                os.setuid(uid)
            log.info("Changed process uid:gid to {}:{}", uid, gid)

        log.info("Started the storage-proxy service.")
    except Exception:
        log.exception("Server initialization failure; triggering shutdown...")
        loop.call_later(0.2, os.kill, 0, signal.SIGINT)
    try:
        yield
    finally:
        log.info("Shutting down...")
        await storage_init_stack.__aexit__(None, None, None)


@click.group(invoke_without_command=True)
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    default=None,
    help=(
        "The config file path. "
        "(default: ./storage-proxy.toml and /etc/backend.ai/storage-proxy.toml)"
    ),
)
@click.option(
    "--debug",
    is_flag=True,
    help="A shortcut to set `--log-level=DEBUG`",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogLevel], case_sensitive=False),
    default=LogLevel.NOTSET,
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    cli_ctx: click.Context,
    config_path: Path,
    log_level: LogLevel,
    debug: bool = False,
) -> int:
    """Start the storage-proxy service as a foreground process."""
    from .watcher import main_job

    log_level = LogLevel.DEBUG if debug else log_level
    try:
        local_config = load_local_config(config_path, log_level=log_level)
    except ConfigurationError as e:
        print(
            "ConfigurationError: Could not read or validate the storage-proxy local config:",
            file=sys.stderr,
        )
        print(pformat(e.invalid_data), file=sys.stderr)
        raise click.Abort()
    # Note: logging configuration is handled separately in Logger class
    # Debug mode is already set during config loading if needed

    multiprocessing.set_start_method("spawn")

    if cli_ctx.invoked_subcommand is None:
        local_config.storage_proxy.pid_file.write_text(str(os.getpid()))
        ipc_base_path = local_config.storage_proxy.ipc_base_path
        log_sockpath = Path(
            ipc_base_path / f"storage-proxy-logger-{os.getpid()}.sock",
        )
        log_sockpath.parent.mkdir(parents=True, exist_ok=True)
        log_endpoint = f"ipc://{log_sockpath}"
        try:
            logger = Logger(
                local_config.logging,
                is_master=True,
                log_endpoint=log_endpoint,
                msgpack_options={
                    "pack_opts": DEFAULT_PACK_OPTS,
                    "unpack_opts": DEFAULT_UNPACK_OPTS,
                },
            )
            with logger:
                setproctitle("backend.ai: storage-proxy")
                log.info("Backend.AI Storage Proxy", VERSION)
                log.info("Runtime: {0}", env_info())
                log.info("Node ID: {0}", local_config.storage_proxy.node_id)
                log_config = logging.getLogger("ai.backend.agent.config")
                if local_config.debug.enabled:
                    log_config.debug("debug mode enabled.")
                if local_config.debug.enabled:
                    print("== Storage proxy configuration ==")
                    pprint(local_config.model_dump())
                match local_config.storage_proxy.event_loop:
                    case EventLoopType.UVLOOP:
                        import uvloop

                        runner = uvloop.run
                        log.info("Using uvloop as the event loop backend")
                    case EventLoopType.ASYNCIO:
                        runner = asyncio.run
                insock_path_prefix = local_config.storage_proxy.watcher_insock_path_prefix
                outsock_path_prefix = local_config.storage_proxy.watcher_outsock_path_prefix
                num_workers = local_config.storage_proxy.num_proc

                if local_config.storage_proxy.use_watcher:
                    if not _is_root():
                        raise InvalidConfigurationSourceError(
                            "Storage proxy must be run as root if watcher is enabled. Else, set"
                            " `use-watcher` to false in your local config file."
                        )
                    insock_path: str | None = local_config.storage_proxy.watcher_insock_path_prefix
                    outsock_path: str | None = (
                        local_config.storage_proxy.watcher_outsock_path_prefix
                    )
                    if insock_path is None or outsock_path is None:
                        raise InvalidSocketPathError(
                            "Socket path must be not null. Please set valid socket path to"
                            " `watcher-insock-path-prefix` and `watcher-outsock-path-prefix` in"
                            " your local config file."
                        )
                    extra_procs = tuple(
                        functools.partial(
                            main_job, worker_pidx, insock_path_prefix, outsock_path_prefix
                        )
                        for worker_pidx in range(num_workers)
                    )
                else:
                    extra_procs = tuple()

                aiotools.start_server(
                    server_main_logwrapper,
                    num_workers=num_workers,
                    extra_procs=extra_procs,
                    args=(local_config, log_endpoint),
                    runner=runner,
                )
                log.info("exit.")
        finally:
            if local_config.storage_proxy.pid_file.is_file():
                # check is_file() to prevent deleting /dev/null!
                local_config.storage_proxy.pid_file.unlink()
    return 0


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, ConfigDict


class ClusterMetricRow(BaseModel):
    model_config = ConfigDict(extra="allow")

    time: int  # timestamp
    total: Optional[int]
    used: Optional[int]
    free: Optional[int]

    def parse(self) -> Optional[ValidClusterMetricRow]:
        if self.total is None or self.used is None or self.free is None:
            return None
        return ValidClusterMetricRow(
            time=self.time,
            total=self.total,
            used=self.used,
            free=self.free,
        )


class ValidClusterMetricRow(BaseModel):
    time: int  # timestamp
    total: int
    used: int
    free: int


class MetricSeries(BaseModel):
    model_config = ConfigDict(extra="allow")

    tags: dict[str, str]
    rows: list[ClusterMetricRow]


class Metric(BaseModel):
    model_config = ConfigDict(extra="allow")

    series: list[MetricSeries]



from __future__ import annotations

from typing import Self, override

from aiohttp import web
from pydantic import ConfigDict

from ai.backend.common.api_handlers import MiddlewareParam

from ..context import RootContext


class StorageRootCtx(MiddlewareParam):
    """Middleware parameter for accessing storage proxy RootContext."""

    root_ctx: RootContext

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    @override
    @classmethod
    async def from_request(cls, request: web.Request) -> Self:
        return cls(root_ctx=request.app["ctx"])



from __future__ import annotations

import abc
import logging
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from ai.backend.common.artifact_storage import AbstractStorage, ImportStepContext
from ai.backend.common.data.artifact.types import ArtifactRegistryType, VerificationStepResult
from ai.backend.common.data.storage.registries.types import FileObjectData
from ai.backend.common.data.storage.types import ArtifactStorageImportStep
from ai.backend.logging import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore


@dataclass
class DownloadStepResult:
    """Result of download step"""

    downloaded_files: list[tuple[FileObjectData, str]]  # (file_info, storage_key)
    storage_name: str
    total_bytes: int


@dataclass
class VerifyStepResult:
    """Result of verify step"""

    verified_files: list[tuple[FileObjectData, str]]  # (file_info, storage_key)
    storage_name: str
    total_bytes: int
    verification_result: VerificationStepResult


InputType = TypeVar("InputType")


class ImportStep(abc.ABC, Generic[InputType]):
    """Base class for import pipeline steps"""

    @property
    @abc.abstractmethod
    def step_type(self) -> ArtifactStorageImportStep:
        """Return the type of this step"""
        pass

    @abc.abstractmethod
    async def execute(self, context: ImportStepContext, input_data: InputType) -> Any:
        """Execute step and return data to pass to next step"""
        pass

    @property
    @abc.abstractmethod
    def registry_type(self) -> ArtifactRegistryType:
        """Return the registry type for this step (used for revision resolution)"""
        pass

    @abc.abstractmethod
    def stage_storage(self, context: ImportStepContext) -> AbstractStorage:
        """Return the storage for this step"""
        pass

    async def cleanup_stage(self, context: ImportStepContext) -> None:
        """Default cleanup implementation that removes files"""
        storage = self.stage_storage(context)
        revision = context.model.resolve_revision(self.registry_type)
        model_prefix = f"{context.model.model_id}/{revision}"

        try:
            await storage.delete_file(model_prefix)
            log.info(f"[cleanup] Removed files: {model_prefix}")
        except Exception as e:
            log.warning(f"[cleanup] Failed to cleanup: {model_prefix}: {str(e)}")


class ImportPipeline:
    """Pipeline that executes import steps in sequence"""

    def __init__(self, steps: list[ImportStep[Any]]) -> None:
        self._steps = steps

    async def execute(self, context: ImportStepContext) -> None:
        """Execute all pipeline steps in sequence"""
        completed_steps: list[ImportStep[Any]] = []
        try:
            current_data: Any = None

            for step in self._steps:
                current_data = await step.execute(context, current_data)
                completed_steps.append(step)

            # On success, cleanup all non-archive steps
            archive_storage = context.storage_step_mappings.get(ArtifactStorageImportStep.ARCHIVE)

            for step in completed_steps:
                if step.step_type != ArtifactStorageImportStep.ARCHIVE:
                    step_storage = context.storage_step_mappings.get(step.step_type)

                    # Skip cleanup if this step uses the same storage as archive step
                    if step_storage == archive_storage:
                        continue

                    try:
                        await step.cleanup_stage(context)
                    except Exception:
                        log.error(f"Failed to cleanup step {step.step_type}")
                        pass

        except Exception:
            # Cleanup completed steps in reverse order on failure
            for step in reversed(completed_steps):
                try:
                    await step.cleanup_stage(context)
                except Exception:
                    # Log cleanup failures but continue with other cleanups
                    log.error(f"Failed to cleanup step {step.step_type}")
                    pass
            raise



from __future__ import annotations

import asyncio
import logging
import uuid
from functools import partial
from typing import Any, Optional

import aiohttp
from pydantic_core._pydantic_core import ValidationError

from ai.backend.common.clients.http_client import ClientKey, ClientPool, tcp_client_session_factory
from ai.backend.common.exception import (
    BaseNFSMountCheckFailed,
    ExportPathNotFound,
    ShowmountNotFound,
)
from ai.backend.common.utils import check_nfs_remote_server
from ai.backend.logging import BraceStyleAdapter

from .exception import (
    AuthenticationError,
    ShareNotFound,
)
from .request import ClusterMetricParams, CreateShareParams, GetShareParams
from .schema.metric import Metric
from .schema.objective import Objective
from .schema.share import Share, SimpleShare
from .schema.site import Site
from .types import APIConnectionInfo

DOMAIN_FOR_HTTP_SESSION = "hammerspace"
NFS_CHECK_RETRY_COUNT = 5
NFS_CHECK_WAIT_SEC = 1

log = BraceStyleAdapter(logging.getLogger(__name__))


class HammerspaceAPIClient:
    def __init__(
        self,
        connection_info: APIConnectionInfo,
    ) -> None:
        custom_session_factory = partial(
            tcp_client_session_factory,
            cookie_jar=aiohttp.CookieJar(unsafe=True),
            ssl=connection_info.ssl_enabled,
        )
        self._client_pool = ClientPool(custom_session_factory)
        self._connection_info = connection_info

    def _create_session(self, key: ClientKey) -> aiohttp.ClientSession:
        session = self._client_pool.load_client_session(key)
        return session

    async def _login(self) -> aiohttp.ClientSession:
        session = self._create_session(
            ClientKey(str(self._connection_info.address), DOMAIN_FOR_HTTP_SESSION)
        )
        async with session.post(
            self._connection_info.address / "login",
            data={
                "username": self._connection_info.username,
                "password": self._connection_info.password,
            },
            ssl=self._connection_info.ssl_enabled,
        ) as resp:
            try:
                resp.raise_for_status()
            except aiohttp.ClientResponseError as e:
                if resp.status // 100 == 4:
                    err_msg = repr(e)
                    raise AuthenticationError(
                        "Hammerspace authentication failed. "
                        f"username: {self._connection_info.username}, err: {err_msg}"
                    ) from e
                raise
        return session

    async def try_login(self) -> None:
        await self._login()

    async def _create_login_session(self) -> aiohttp.ClientSession:
        session = await self._login()
        return session

    async def _check_nfs_export(self, export_path: str) -> None:
        host = self._connection_info.address.host
        if host is None:
            log.warning(
                "Invalid host in the address: {}, skip NFS export check",
                self._connection_info.address,
            )
            return
        count = NFS_CHECK_RETRY_COUNT
        while count > 0:
            count -= 1
            try:
                await check_nfs_remote_server(host, export_path)
            except ShowmountNotFound:
                log.warning("showmount command not found. Install nfs-common package.")
                return
            except ExportPathNotFound:
                log.debug(
                    "NFS export path {} not found on the server {}. Retrying...", export_path, host
                )
            except BaseNFSMountCheckFailed:
                log.exception(
                    "Check NFS export {} on the server {} failed. Retrying...", export_path, host
                )
            else:
                log.debug("Found NFS export {} on the server {}", export_path, host)
                return
            await asyncio.sleep(NFS_CHECK_WAIT_SEC)
        log.exception("Failed to verify NFS export {} on the server {}", export_path, host)

    async def create_share(
        self,
        params: CreateShareParams,
        retry: int,
        wait_sec: int,
    ) -> SimpleShare:
        session = await self._create_login_session()
        async with session.post(
            self._connection_info.address / "shares",
            params=params.query(),
            json=params.body(),
            ssl=self._connection_info.ssl_enabled,
        ) as resp:
            resp.raise_for_status()

        while retry > 0:
            log.debug(
                "Retrying to get the created share: {}, remaining retries: {}", params.name, retry
            )
            raw_shares = await self._get_shares(session, GetShareParams(name=params.name))
            for raw_share in raw_shares:
                try:
                    share = SimpleShare.model_validate(raw_share)
                    if share.name == params.name:
                        await self._check_nfs_export(share.path)
                        return share
                except ValidationError:
                    continue
            retry -= 1
            await asyncio.sleep(wait_sec)  # wait for a moment and try again
        raise ShareNotFound(f"Failed to get the created share: {params.name}")

    async def _get_shares(
        self, session: aiohttp.ClientSession, params: GetShareParams
    ) -> list[dict[str, Any]]:
        session = await self._create_login_session()
        async with session.get(
            self._connection_info.address / "shares",
            params=params.query(),
            ssl=self._connection_info.ssl_enabled,
        ) as resp:
            resp.raise_for_status()
            data = await resp.json()
            return data

    async def get_share(
        self,
        params: GetShareParams,
    ) -> Optional[Share]:
        session = await self._create_login_session()
        data = await self._get_shares(session, params)
        for raw_share in data:
            try:
                share = Share.model_validate(raw_share)
                if share.name == params.name:
                    return share
            except ValidationError:
                continue
        return None

    async def poll_share(
        self,
        params: GetShareParams,
        retry: int,
        wait_sec: int,
    ) -> Optional[Share]:
        session = await self._create_login_session()
        while retry > 0:
            data = await self._get_shares(session, params)
            for raw_share in data:
                try:
                    share = Share.model_validate(raw_share)
                except ValidationError:
                    continue
                if share.name == params.name:
                    return share
            retry -= 1
            await asyncio.sleep(wait_sec)  # wait for a moment and try again
        return None

    async def get_objective(self, id: uuid.UUID) -> Optional[Objective]:
        session = await self._create_login_session()
        query = {"spec": f"uoid.uuid=eq={id}"}
        async with session.get(self._connection_info.address / "objectives", params=query) as resp:
            resp.raise_for_status()
            data = await resp.json()
            objectives = [Objective.model_validate(obj) for obj in data]
            for obj in objectives:
                if obj.uoid.uuid == id:
                    return obj
            return None

    async def get_sites(self) -> list[Site]:
        session = await self._create_login_session()
        async with session.get(self._connection_info.address / "sites") as resp:
            resp.raise_for_status()
            data = await resp.json()
            sites: list[Site] = []
            for s in data:
                try:
                    site = Site.model_validate(s)
                    sites.append(site)
                except ValidationError:
                    log.warning("Failed to parse the site data: {}", s)
                    continue
            return sites

    async def get_cluster_metrics(self, params: ClusterMetricParams) -> Metric:
        session = await self._create_login_session()
        async with session.get(
            self._connection_info.address
            / f"metrics/capacity/{params.object_type}/{params.site_id}",
            params=params.query(),
        ) as resp:
            resp.raise_for_status()
            data = await resp.json()
            metric = Metric.model_validate(data)
            return metric



from dataclasses import dataclass
from typing import Optional

from pydantic import BaseModel
from yarl import URL


class SSLConfig(BaseModel):
    cert_file: Optional[str]
    key_file: Optional[str]


@dataclass
class APIConnectionInfo:
    address: URL
    username: str
    password: str
    ssl_enabled: bool
    ssl_config: Optional[SSLConfig]



import enum


class ObjectType(enum.StrEnum):
    NONE = "NONE"
    ANY = "ANY"
    EVENT = "EVENT"
    MDS = "MDS"
    NODE = "NODE"
    FAN = "FAN"
    NIC = "NIC"
    HDD = "HDD"
    SSD = "SSD"
    DISK_DRIVE = "DISK_DRIVE"
    POWER_SUPPLY = "POWER_SUPPLY"
    DATA_SPHERE = "DATA_SPHERE"
    DATA_MOVER = "DATA_MOVER"
    DATA_PORTAL = "DATA_PORTAL"
    STORAGE_VOLUME = "STORAGE_VOLUME"
    VOLUME_GROUP = "VOLUME_GROUP"
    NETWORK_IF = "NETWORK_IF"
    MGMT = "MGMT"
    SHARE = "SHARE"
    BASE_USER = "BASE_USER"
    USER = "USER"
    FEDERATED_USER = "FEDERATED_USER"
    USER_GROUP = "USER_GROUP"
    IDENTITY_GROUP_MAPPING = "IDENTITY_GROUP_MAPPING"
    IDENTITY = "IDENTITY"
    MANAGEMENT_ROLE = "MANAGEMENT_ROLE"
    STEP = "STEP"
    TASK = "TASK"
    SMTP = "SMTP"
    NOTIFICATION_RULE = "NOTIFICATION_RULE"
    HEARTBEAT = "HEARTBEAT"
    NTP = "NTP"
    LOGICAL_VOLUME = "LOGICAL_VOLUME"
    LDAP = "LDAP"
    IDP = "IDP"
    DNS = "DNS"
    GATEWAY = "GATEWAY"
    DHCP = "DHCP"
    CLUSTER = "CLUSTER"
    CPU = "CPU"
    ELEMENTAL_OBJECTIVE = "ELEMENTAL_OBJECTIVE"
    OBJECTIVE = "OBJECTIVE"
    DATA_MANAGER = "DATA_MANAGER"
    SCHEDULE = "SCHEDULE"
    CONDITION = "CONDITION"
    SNAPSHOT_RETENTION = "SNAPSHOT_RETENTION"
    FILE_SNAPSHOT = "FILE_SNAPSHOT"
    OBJECT_STORE = "OBJECT_STORE"
    OBJECT_STORAGE_VOLUME = "OBJECT_STORAGE_VOLUME"
    CLOUD_MOVER = "CLOUD_MOVER"
    OBJECT_STORE_LOGICAL_VOLUME = "OBJECT_STORE_LOGICAL_VOLUME"
    NODE_LOCATION = "NODE_LOCATION"
    VOLUME_LOCATION = "VOLUME_LOCATION"
    SNMP = "SNMP"
    SW_UPDATE_TASK = "SW_UPDATE_TASK"
    CTDB = "CTDB"
    SHARE_SNAPSHOT = "SHARE_SNAPSHOT"
    BACKUP = "BACKUP"
    SAMBA_AD = "SAMBA_AD"
    DOMAIN_IDMAP = "DOMAIN_IDMAP"
    HMDB = "HMDB"
    SYSLOG = "SYSLOG"
    NIS = "NIS"
    NIS_CONFIG = "NIS_CONFIG"
    SUBNET_GATEWAY = "SUBNET_GATEWAY"
    STATIC_ROUTE = "STATIC_ROUTE"
    SITE = "SITE"
    SHARE_PARTICIPANT = "SHARE_PARTICIPANT"
    LOGIN_POLICY = "LOGIN_POLICY"
    DATA_COPY_TO_OBJECT_TASK = "DATA_COPY_TO_OBJECT_TASK"
    KMS = "KMS"
    BASE_STORAGE_VOLUME = "BASE_STORAGE_VOLUME"
    LICENSE_ACTIVATION = "LICENSE_ACTIVATION"
    SMB_ALIAS = "SMB_ALIAS"
    LOGICAL_SHARE = "LOGICAL_SHARE"
    WORM_ENTRY = "WORM_ENTRY"
    ANTIVIRUS = "ANTIVIRUS"
    ALCHEMI = "ALCHEMI"
    LABEL = "LABEL"
    S3_SERVER_CONFIG = "S3_SERVER_CONFIG"
    S3_SERVER_USER = "S3_SERVER_USER"
    S3_SERVER_BUCKET = "S3_SERVER_BUCKET"
    S3_SERVER_BUCKET_CONTAINER = "S3_SERVER_BUCKET_CONTAINER"



import logging

from ai.backend.common.dto.storage.response import (
    ObjectMetaResponse,
    PresignedDownloadObjectResponse,
    PresignedUploadObjectResponse,
)
from ai.backend.common.types import StreamReader
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.storage.errors import (
    ObjectStorageBucketNotFoundError,
    StorageNotFoundError,
    StorageTypeInvalidError,
)
from ai.backend.storage.storages.object_storage import ObjectStorage
from ai.backend.storage.storages.storage_pool import StoragePool

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ObjectStorageService:
    """
    Service class for S3 storage operations.
    """

    _storage_pool: StoragePool

    def __init__(self, storage_pool: StoragePool) -> None:
        self._storage_pool = storage_pool

    async def stream_upload(
        self,
        storage_name: str,
        bucket_name: str,
        filepath: str,
        data_stream: StreamReader,
    ) -> None:
        """
        Upload a file to S3 using streaming.

        Args:
            storage_name: Name of the storage configuration
            bucket_name: Name of the S3 bucket
            filepath: Path to the file to upload
            content_type: Content type of the file
            data_stream: Async iterator of file data chunks
        """
        storage = self._resolve_storage(storage_name, bucket_name)
        await storage.stream_upload(filepath, data_stream)

    async def stream_download(
        self, storage_name: str, bucket_name: str, filepath: str
    ) -> StreamReader:
        """
        Download a file from S3 using streaming.

        Args:
            storage_name: Name of the storage configuration
            bucket_name: Name of the S3 bucket
            filepath: Path to the file to download

        Returns:
            FileStream: Stream for reading file data
        """
        storage = self._resolve_storage(storage_name, bucket_name)
        return await storage.stream_download(filepath)

    async def get_object_info(
        self, storage_name: str, bucket_name: str, filepath: str
    ) -> ObjectMetaResponse:
        """
        Get object information.

        Args:
            storage_name: Name of the storage configuration
            bucket_name: Name of the S3 bucket
            filepath: Path to the file to download

        Returns:
            ObjectMetaResponse with object metadata
        """
        storage = self._resolve_storage(storage_name, bucket_name)
        return await storage.get_file_info(filepath)

    async def delete_object(self, storage_name: str, bucket_name: str, prefix: str) -> None:
        """
        Delete an object and all its contents.

        Args:
            storage_name: Name of the storage configuration
            bucket_name: Name of the S3 bucket
            prefix: Prefix of the object to delete
        """
        storage = self._resolve_storage(storage_name, bucket_name)
        await storage.delete_file(prefix)

    async def generate_presigned_upload_url(
        self, storage_name: str, bucket_name: str, key: str
    ) -> PresignedUploadObjectResponse:
        """
        Generate presigned upload URL.

        Args:
            storage_name: Name of the storage configuration
            bucket_name: Name of the S3 bucket
            key: Path to the file to upload

        Returns:
            PresignedUploadObjectResponse with URL and fields
        """
        storage = self._resolve_storage(storage_name, bucket_name)
        return await storage.generate_presigned_upload_url(key)

    async def generate_presigned_download_url(
        self,
        storage_name: str,
        bucket_name: str,
        filepath: str,
    ) -> PresignedDownloadObjectResponse:
        """
        Generate presigned download URL.

        Args:
            storage_name: Name of the storage configuration
            bucket_name: Name of the S3 bucket
            filepath: Path to the file to download

        Returns:
            PresignedDownloadResponse with URL
        """
        storage = self._resolve_storage(storage_name, bucket_name)
        return await storage.generate_presigned_download_url(filepath)

    def _resolve_storage(self, storage_name: str, bucket_name: str) -> ObjectStorage:
        try:
            storage = self._storage_pool.get_storage(storage_name)
            if not isinstance(storage, ObjectStorage):
                raise StorageTypeInvalidError(f"Storage '{storage_name}' is not an ObjectStorage")

            # TODO: Remove this after supporting multiple buckets
            if storage._bucket != bucket_name:
                raise ObjectStorageBucketNotFoundError(
                    f"Bucket '{bucket_name}' not configured for storage '{storage_name}'. "
                    f"Expected bucket: '{storage._bucket}'"
                )

            return storage
        except KeyError:
            raise StorageNotFoundError(
                f"No storage configuration found for storage: {storage_name}"
            )



import enum

from pydantic import BaseModel, ConfigDict

from .df_entry import DFEntry
from .objective import SimpleObjective
from .uoid import UOID


class ShareState(enum.StrEnum):
    OFFLINE = "OFFLINE"
    ONLINE = "ONLINE"
    MOUNTED = "MOUNTED"
    PUBLISHED = "PUBLISHED"
    CREATED = "CREATED"
    PRE_REMOVAL = "PRE_REMOVAL"
    REMOVING = "REMOVING"
    REMOVED = "REMOVED"
    RECLAIMING = "RECLAIMING"
    REMOVE_FAILED = "REMOVE_FAILED"


class ShareLifecycle(enum.StrEnum):
    CREATING = "CREATING"
    CREATED = "CREATED"
    PRE_DELETE = "PRE_DELETE"
    DELETE_SCHEDULED = "DELETE_SCHEDULED"
    DELETING = "DELETING"
    DELETE_FAILED = "DELETE_FAILED"


class ShareObjective(BaseModel):
    model_config = ConfigDict(extra="allow")

    objective: SimpleObjective
    applicability: str
    removable: bool


class SimpleShare(BaseModel):
    model_config = ConfigDict(extra="allow")

    uoid: UOID
    name: str
    path: str


class Share(BaseModel):
    model_config = ConfigDict(extra="allow")

    uoid: UOID
    name: str
    modified: int  # timestamp
    totalNumberOfFiles: int
    inodes: DFEntry
    space: DFEntry
    path: str
    shareState: ShareState
    shareLifecycle: ShareLifecycle
    shareObjectives: list[ShareObjective]



import asyncio
import logging
from pathlib import Path
from subprocess import CalledProcessError
from typing import Any, Final, FrozenSet, Mapping

import aiofiles
import aiofiles.os

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.types import QuotaScopeID
from ai.backend.logging import BraceStyleAdapter

from ...errors import (
    DDNCommandFailedError,
    QuotaScopeAlreadyExists,
    QuotaScopeNotFoundError,
    SubprocessStdoutNotAvailableError,
)
from ...subproc import run
from ...types import Optional, QuotaConfig, QuotaUsage
from ..abc import CAP_QUOTA, CAP_VFOLDER, AbstractQuotaModel
from ..vfs import BaseQuotaModel, BaseVolume

FIRST_PROJECT_ID: Final = 100
PROJECT_MAIN_ID_KEY: Final = "ddn/main-project-id"
PROJECT_ID_FILE_NAME: Final = "project_id"


def _byte_to_kilobyte(byte: int) -> int:
    return byte // 1024


def _kilobyte_to_byte(kilobyte: int) -> int:
    return kilobyte * 1024


log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class EXAScalerQuotaModel(BaseQuotaModel):
    def __init__(self, mount_path: Path, local_config: Mapping[str, Any], etcd: AsyncEtcd) -> None:
        self.local_config = local_config
        self.etcd = etcd
        super().__init__(mount_path)
        return

    async def _read_project_id(self, pid_file_path: str | Path) -> int | None:
        def _read():
            try:
                with open(pid_file_path, "r") as f:
                    return int(f.read())
            except FileNotFoundError:
                return None

        return await asyncio.get_running_loop().run_in_executor(None, _read)

    async def _write_project_id(self, pid: int, pid_file_path: str | Path) -> None:
        def _write():
            with open(pid_file_path, "w") as f:
                f.write(str(pid))

        await asyncio.get_running_loop().run_in_executor(None, _write)

    async def _read_main_project_id(self) -> int:
        raw_val = await self.etcd.get(PROJECT_MAIN_ID_KEY)
        if raw_val is None:
            val = int(FIRST_PROJECT_ID)
        else:
            val = int(raw_val)
        await self.etcd.put(PROJECT_MAIN_ID_KEY, str(val + 1))
        return val

    async def _set_quota_by_project(self, pid: int, path: Path, options: QuotaConfig) -> None:
        # default unit for DDN quota is KB
        quota_limit = _byte_to_kilobyte(options.limit_bytes)
        try:
            await run([
                b"sudo",
                b"lfs",
                b"setquota",
                b"-p",
                str(pid),
                f"-B{quota_limit}",
                path,
            ])
        except CalledProcessError as e:
            raise DDNCommandFailedError(f"'lfs setquota -p {pid}' command failed: {e.stderr}")

    async def _unset_quota_by_project(self, pid: int, path: Path) -> None:
        await self._set_quota_by_project(pid, path, QuotaConfig(0))

    async def _get_quota_by_project(self, pid: int, qspath: Path) -> QuotaUsage | None:
        proc = await asyncio.create_subprocess_exec(
            b"lfs",
            b"quota",
            b"-p",
            str(pid),
            str(qspath),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )
        try:
            if proc.stdout is None:
                raise SubprocessStdoutNotAvailableError("lfs quota process stdout is not available")
            next_line_is_quota = False
            while True:
                try:
                    raw = await proc.stdout.readline()
                    if not raw:
                        break
                    line = raw.decode()
                except asyncio.IncompleteReadError:
                    break
                words = line.split()
                if next_line_is_quota:
                    raw_used_bytes, hard_limit = words[0], int(words[2])
                    # words[1] is soft_limit
                    if hard_limit == 0:
                        return None
                    if raw_used_bytes.endswith("*"):
                        raw_used_bytes = raw_used_bytes[:-1]
                    used_bytes = _kilobyte_to_byte(int(raw_used_bytes))
                    limit_bytes = _kilobyte_to_byte(hard_limit)
                    if used_bytes < 0 or limit_bytes < 0:
                        log.warning(
                            "Subprocess lfs quota returned negative values in used_bytes({}) or limit_bytes({}), pid: {} Line: {}",
                            used_bytes,
                            limit_bytes,
                            pid,
                            line,
                        )
                    return QuotaUsage(used_bytes=used_bytes, limit_bytes=limit_bytes)
                if Path(words[0]) == qspath:
                    next_line_is_quota = True
                    continue
            return None
        finally:
            await proc.wait()

    async def create_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        options: Optional[QuotaConfig] = None,
        extra_args: Optional[dict[str, Any]] = None,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        pid_path = qspath / PROJECT_ID_FILE_NAME
        try:
            await aiofiles.os.makedirs(qspath)
        except FileExistsError:
            pass
        project_id = await self._read_project_id(pid_path)
        if project_id is None:
            main_pid = await self._read_main_project_id()
            project_id = main_pid + 1
            await self._write_project_id(project_id, pid_path)
        else:
            quota_usage = await self._get_quota_by_project(project_id, qspath)
            if quota_usage is not None:
                raise QuotaScopeAlreadyExists

        # Set projectID to the directory
        try:
            await run([
                b"sudo",
                b"lfs",
                b"project",
                b"-p",
                str(project_id),
                b"-r",
                b"-s",
                str(qspath),
            ])
        except CalledProcessError as e:
            raise DDNCommandFailedError(f"'lfs project -p {project_id}' command failed: {e.stderr}")

        if options is not None:
            await self._set_quota_by_project(project_id, qspath, options)

    async def describe_quota_scope(self, quota_scope_id: QuotaScopeID) -> QuotaUsage | None:
        """
        $ lfs quota -p <projectId> <fs_mount_point>

        Disk quotas for prj <PID> (pid <PID>):
        Filesystem  kbytes   quota   limit   grace   files   quota   limit   grace
        /mnt/lufs01/vfroot/test
                    1004       0       2048       -       2       0       0       -
        pid <PID> is using default file quota setting

        ---

        `kbytes` is quota usage. `quota` is soft limit and `limit` is hard limit.
        It will remove files after the `grace` if you exceed soft limit.
        """

        qspath = self.mangle_qspath(quota_scope_id)
        if not qspath.exists():
            return None
        pid_path = qspath / PROJECT_ID_FILE_NAME
        if (pid := await self._read_project_id(pid_path)) is None:
            return None

        return await self._get_quota_by_project(pid, qspath)

    async def update_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        config: QuotaConfig,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        pid_path = qspath / PROJECT_ID_FILE_NAME
        pid = await self._read_project_id(pid_path)
        if pid is None:
            raise QuotaScopeNotFoundError
        await self._set_quota_by_project(pid, qspath, config)

    async def unset_quota(self, quota_scope_id: QuotaScopeID) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        pid_path = qspath / PROJECT_ID_FILE_NAME
        pid = await self._read_project_id(pid_path)
        if pid is None:
            raise QuotaScopeNotFoundError
        await self._unset_quota_by_project(pid, qspath)

    async def delete_quota_scope(self, quota_scope_id: QuotaScopeID) -> None:
        await self.unset_quota(quota_scope_id)
        qspath = self.mangle_qspath(quota_scope_id)
        await aiofiles.os.rmdir(qspath)


class EXAScalerFSVolume(BaseVolume):
    name = "exascaler"

    async def create_quota_model(self) -> AbstractQuotaModel:
        return EXAScalerQuotaModel(self.mount_path, self.local_config, self.etcd)

    async def get_capabilities(self) -> FrozenSet[str]:
        return frozenset([CAP_VFOLDER, CAP_QUOTA])
